"""
The translation from Abstract Syntax Tree to Python happens in both ways.
The outer expression is visited with its builder, which will split its
arguments and visit them with their respective builders. Once the lowest
level is reached, it will be translated into Python returning a BuildAST
object, this object will include the python expression, its subscripts,
its calls to other and its arithmetic order (see Build AST for more info).
BuildAST will be returned for each visited argument from the lower
lever to the top level, giving the final expression.
"""
import warnings
from dataclasses import dataclass
from typing import Union

import numpy as np
from pysd.py_backend.utils import compute_shape

from pysd.translators.structures.abstract_expressions import\
    AbstractSyntax, AllocateAvailableStructure, AllocateByPriorityStructure,\
    ArithmeticStructure, CallStructure, DataStructure, DelayFixedStructure,\
    DelayStructure, DelayNStructure, ForecastStructure, GameStructure,\
    GetConstantsStructure, GetDataStructure, GetLookupsStructure,\
    InitialStructure, InlineLookupsStructure, IntegStructure,\
    LogicStructure, LookupsStructure, ReferenceStructure,\
    SampleIfTrueStructure, SmoothNStructure, SmoothStructure,\
    SubscriptsReferenceStructure, TrendStructure

from .python_functions import functionspace
from .subscripts import SubscriptManager


@dataclass
class BuildAST:
    """
    Python expression holder.

    Parameters
    ----------
    expression: str
        The Python expression.
    calls: dict
        The calls to other variables for the dependencies dictionary.
    subscripts: dict
        The subscripts dict of the expression.
    order: int
        Arithmetic order of the expression. The arithmetic order depends
        on the last arithmetic operation. If the expression is a number,
        a call to a function, or is between parenthesis; its order will
        be 0. If the expression its an exponential of two terms its order
        will be 1. If the expression is a product or division its order
        will be 2. If the expression is a sum or substraction its order
        will be 3. If the expression is a logical comparison its order
        will be 4.

    """
    expression: str
    calls: dict
    subscripts: dict
    order: int

    def __str__(self) -> str:
        # makes easier building
        return self.expression

    def reshape(self, subscripts: SubscriptManager,
                final_subscripts: dict,
                final_element: bool = False) -> None:
        """
        Reshape the object to the desired subscripts. It will modify the
        expression and lower the order if it is not 0.

        Parameters
        ----------
        subscripts: SubscriptManager
            The subscripts of the section.
        final_subscripts: dict
            The desired final subscripts.
        final_element: bool (optional)
            If True the array will be reshaped with the final subscripts
            to have the shame shape. Otherwise, a length 1 dimension
            will be included in the position to allow arithmetic
            operations with other arrays. Default is False.

        """
        if not final_subscripts or (
          self.subscripts == final_subscripts
          and list(self.subscripts) == list(final_subscripts)):
            # Same dictionary in the same order, do nothing
            pass
        elif not self.subscripts:
            # Original expression is not an array
            # NUMPY: object.expression = np.full(%s, %(shape)s)
            subscripts_out = subscripts.simplify_subscript_input(
                final_subscripts)[1]
            self.expression = "xr.DataArray(%s, %s, %s)" % (
                self.expression, subscripts_out, list(final_subscripts)
            )
            self.order = 0
            self.subscripts = final_subscripts
        else:
            # Original expression is an array
            self.lower_order(-1)

            # Reorder subscrips
            final_order = {
                sub: self.subscripts[sub]
                for sub in final_subscripts
                if sub in self.subscripts
            }
            if list(final_order) != list(self.subscripts):
                # NUMPY: reorder dims if neccessary with np.moveaxis or similar
                self.expression +=\
                    f".transpose({', '.join(map(repr, final_order))})"
                self.subscripts = final_order

            # add new dimensions
            if final_element and final_subscripts != self.subscripts:
                # NUMPY: remove final_element condition from top
                # NUMPY: add new axis with [:, None, :]
                # NUMPY: move final_element condition here and use np.tile
                for i, dim in enumerate(final_subscripts):
                    if dim not in self.subscripts:
                        subscripts_out = subscripts.simplify_subscript_input(
                            {dim: final_subscripts[dim]})[1]
                        self.expression +=\
                            f".expand_dims({subscripts_out}, {i})"

                self.subscripts = final_subscripts

    def lower_order(self, new_order: int) -> None:
        """
        Lower the order to maintain the correct order in arithmetic
        operations. If the requested order is smaller than the current
        order parenthesis will be added to the expression to lower its
        order to 0.

        Parameters
        ----------
        new_order: int
            The required new order of the expression. If 0 it will be
            assumed that the expression will be passed as an argument
            of a function and therefore no operations will be done. If
            order 0 is required, a negative value can be used for
            new_order.

        """
        if self.order >= new_order and self.order != 0 and new_order != 0:
            # if current operator order is 0 do not need to do anything
            # if the order of operations conflicts add parenthesis
            # if new order is 0 do not need to do anything, as it may be
            # an argument to a function. To force the 0 order a negative
            # value can be used, which will force the parenthesis
            # (necessary to reshape some arrays)
            self.expression = "(%s)" % self.expression
            self.order = 0


class StructureBuilder:
    """
    Main builder for Abstract Syntax Tree structures. All the builders
    are children of this class, which allows them inheriting the methods.
    """
    def __init__(self, value: object, component: object):
        # component typing should be ComponentBuilder, but importing it
        # for typing would create a circular dependency :S
        self.value = value
        self.arguments = {}
        self.component = component
        self.element = component.element
        self.section = component.section
        self.def_subs = component.subscripts_dict

    @staticmethod
    def join_calls(arguments: dict) -> dict:
        """
        Merge the calls of the arguments.

        Parameters
        ----------
        arguments: dict
            The dictionary of arguments. The keys should br strings of
            ordered integer numbers starting from 0.

        Returns
        -------
        calls: dict
            The merged dictionary of calls.

        """
        if len(arguments) == 0:
            # No arguments
            return {}
        elif len(arguments) == 1:
            # Only one argument
            return list(arguments.values())[0].calls
        else:
            # Several arguments
            return merge_dependencies(
                *[val.calls for val in arguments.values()])

    def reorder(self, arguments: dict, force: bool = None) -> dict:
        """
        Reorder the subscripts of the arguments to make them match.

        Parameters
        ----------
        arguments: dict
            The dictionary of arguments. The keys should br strings of
            ordered integer numbers starting from 0.
        force: 'component', 'equal', or None (optional)
            If force is 'component' it will force the arguments to have
            the subscripts of the component definition. If force is
            'equal' it will force all the arguments to have the same
            subscripts, includying the floats. If force is None, it
            will only modify the shape of the arrays adding length 1
            dimensions to allow operation between different shape arrays.
            Default is None.

        Returns
        -------
        final_subscripts: dict
            The final_subscripts after reordering all the elements.

        """
        if force == "component":
            final_subscripts = self.def_subs or {}
        else:
            final_subscripts = self.get_final_subscripts(arguments)

        [arguments[key].reshape(
            self.section.subscripts, final_subscripts, bool(force))
         for key in arguments
         if arguments[key].subscripts or force == "equal"]

        return final_subscripts

    def get_final_subscripts(self, arguments: dict) -> dict:
        """
        Get the final subscripts of a combination of arguments.

        Parameters
        ----------
        arguments: dict
            The dictionary of arguments. The keys should br strings of
            ordered integer numbers starting from 0.

        Returns
        -------
        final_subscripts: dict
            The final_subscripts of combining all the elements.

        """
        if len(arguments) == 0:
            return {}
        elif len(arguments) == 1:
            return arguments["0"].subscripts
        else:
            return self._compute_final_subscripts(
                [arg.subscripts for arg in arguments.values()])

    def _compute_final_subscripts(self, subscripts_list: list) -> dict:
        """
        Compute final subscripts from a list of subscript dictionaries.

        Parameters
        ----------
        subscript_list: list of dicts
            List of subscript dictionaries.

        """
        expression = {}
        [expression.update(subscript)
         for subscript in subscripts_list if subscript]
        # TODO reorder final_subscripts taking into account def_subs
        # this way try to minimize the reordering operations
        return expression

    def update_object_subscripts(self, name: str,
                                 component_final_subs: dict) -> None:
        """
        Update the object subscripts. Needed for those objects that
        use 'add' method to load several components at once and mixed
        definitions are used.

        Parameters
        ----------
        name: str
            The name of the object in the objects dictionary from the
            element.
        component_final_subs: dict
            The subscripts of the component but with the element
            subscript ranges as keys. This can differ from the component
            subscripts when the component is defined with subranges of
            the final subscript ranges.

        """
        # Get the component used to define the object first time
        origin_comp = self.element.objects[name]["component"]
        # The original component subscript dictionary is a list
        origin_comp.subscripts_dict.append(component_final_subs)


class OperationBuilder(StructureBuilder):
    """Builder for arithmetic and logical operations."""
    _operators_build = {
        "^": ("%(left)s**%(right)s", None, 1),
        "*": ("%(left)s*%(right)s", None, 2),
        "/": ("%(left)s/%(right)s", None, 2),
        "+": ("%(left)s + %(right)s", None, 3),
        "-": ("%(left)s - %(right)s", None, 3),
        "=": ("%(left)s == %(right)s", None, 4),
        "<>": ("%(left)s != %(right)s", None, 4),
        ">=": ("%(left)s >= %(right)s", None, 4),
        ">": ("%(left)s > %(right)s", None, 4),
        "<=": ("%(left)s <= %(right)s", None, 4),
        "<": ("%(left)s < %(right)s", None, 4),
        ":NOT:": ("np.logical_not(%s)", ("numpy",), 0),
        ":AND:": ("np.logical_and(%(left)s, %(right)s)", ("numpy",), 0),
        ":OR:": ("np.logical_or(%(left)s, %(right)s)", ("numpy",), 0),
        "negative": ("-%s", None, 3),
    }

    def __init__(self, operation: Union[ArithmeticStructure, LogicStructure],
                 component: object):
        super().__init__(None, component)
        self.operators = operation.operators.copy()
        self.arguments = {
            str(i): arg for i, arg in enumerate(operation.arguments)}

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        operands = {}
        calls = self.join_calls(arguments)
        final_subscripts = self.reorder(arguments)
        arguments = [arguments[str(i)] for i in range(len(arguments))]
        dependencies, order = self._operators_build[self.operators[-1]][1:]

        if dependencies:
            # Add necessary dependencies to the imports
            self.section.imports.add(*dependencies)

        if self.operators[-1] == "^":
            # Right side of the exponential can be from higher order
            arguments[-1].lower_order(2)
        else:
            arguments[-1].lower_order(order)

        if len(arguments) == 1:
            # not and negative operations (only 1 element)
            if self.operators[0] == "negative":
                order = 1
            expression = self._operators_build[self.operators[0]][0]
            return BuildAST(
                expression=expression % arguments[0],
                calls=calls,
                subscripts=final_subscripts,
                order=order)

        # Add the arguments to the expression with the operator,
        # they are built from right to left
        # Get the last argument as the RHS of the first operation
        operands["right"] = arguments.pop()
        while arguments or self.operators:
            # Get the operator and the LHS of the operation
            expression = self._operators_build[self.operators.pop()][0]
            operands["left"] = arguments.pop()
            # Lower the order of the LHS if neccessary
            operands["left"].lower_order(order)
            # Include the operation in the RHS for next iteration
            operands["right"] = expression % operands

        return BuildAST(
            expression=operands["right"],
            calls=calls,
            subscripts=final_subscripts,
            order=order)


class GameBuilder(StructureBuilder):
    """Builder for GAME expressions."""
    def __init__(self, game_str: GameStructure, component: object):
        super().__init__(None, component)
        self.arguments = {"expr": game_str.expression}

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        # Game calls are ignored as we have no support for a similar
        # feature, we simply return the content inside the GAME call
        return arguments["expr"]


class CallBuilder(StructureBuilder):
    """Builder for calls to functions, macros and lookups."""
    def __init__(self, call_str: CallStructure, component: object):
        super().__init__(None, component)
        function_name = call_str.function.reference
        self.arguments = {
            str(i): arg for i, arg in enumerate(call_str.arguments)}

        if function_name in self.section.macrospace:
            # Build macro
            self.macro_name = function_name
            self.build = self.build_macro_call
        elif function_name in self.section.namespace.cleanspace:
            # Build lookupcall
            self.arguments["function"] = call_str.function
            self.build = self.build_lookups_call
        elif function_name in functionspace:
            # Build direct function
            self.function = function_name
            self.build = self.build_function_call
        elif function_name == "a_function_of":
            # Build incomplete function
            self.build = self.build_incomplete_call
        else:
            # Build missing function
            self.function = function_name
            self.build = self.build_not_implemented

    def build_not_implemented(self, arguments: dict) -> BuildAST:
        """
        Build method for not implemented function calls.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        final_subscripts = self.reorder(arguments)
        warnings.warn(
            "Trying to translate '" + self.function.upper().replace("_", " ")
            + "' which it is not implemented on PySD. The translated "
            + "model will crash..."
        )
        self.section.imports.add("functions", "not_implemented_function")

        return BuildAST(
            expression="not_implemented_function('%s', %s)" % (
                self.function,
                ", ".join(arg.expression for arg in arguments.values())),
            calls=self.join_calls(arguments),
            subscripts=final_subscripts,
            order=0)

    def build_incomplete_call(self, arguments: dict) -> BuildAST:
        """
        Build method for incomplete function calls.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        warnings.warn(
            "'%s' has no equation specified" % self.element.name,
            SyntaxWarning, stacklevel=2
        )
        self.section.imports.add("functions", "incomplete")
        return BuildAST(
            expression="incomplete(%s)" % ", ".join(
                arg.expression for arg in arguments.values()),
            calls=self.join_calls(arguments),
            subscripts=self.def_subs,
            order=0)

    def build_macro_call(self, arguments: dict) -> BuildAST:
        """
        Build method for macro calls.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.section.imports.add("model", "Macro")
        # Get macro from macrospace
        macro = self.section.macrospace[self.macro_name]

        calls = self.join_calls(arguments)
        final_subscripts = self.reorder(arguments)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.macro_name + "_" + self.element.identifier, prefix="_macro")
        arguments["file"] = macro.path.name
        arguments["macro_name"] = macro.name
        arguments["args"] = "{%s}" % ", ".join([
            "'%s': lambda: %s" % (key, val)
            for key, val in zip(macro.params, arguments.values())
        ])

        # Create Macro object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = Macro(_root.joinpath('%(file)s'), "
                          "%(args)s, '%(macro_name)s', "
                          "time_initialization=lambda: __data['time'], "
                          "py_name='%(name)s')" % arguments,
        }
        # Add other_dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": calls,
            "step": calls
        }

        return BuildAST(
            expression="%s()" % arguments["name"],
            calls={arguments["name"]: 1},
            subscripts=final_subscripts,
            order=0)

    def build_lookups_call(self, arguments: dict) -> BuildAST:
        """
        Build method for loookups calls.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        if arguments["0"].subscripts:
            # Build lookups with subcripted arguments
            # it is neccessary to give the final subscripts information
            # in the call to rearrange it correctly
            final_subscripts =\
                self.get_final_subscripts(arguments)
            expression = arguments["function"].expression.replace(
                "()", f"(%(0)s, {final_subscripts})")
        else:
            # Build lookups with float arguments
            final_subscripts = arguments["function"].subscripts
            expression = arguments["function"].expression.replace(
                "()", "(%(0)s)")

        # NUMPY: we need to manage inside lookup with subscript and later
        #        return the values in a correct ndarray
        return BuildAST(
            expression=expression % arguments,
            calls=self.join_calls(arguments),
            subscripts=final_subscripts,
            order=0)

    def build_function_call(self, arguments: dict) -> BuildAST:
        """
        Build method for function calls.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        # Get the function expression from the functionspace
        expression, modules = functionspace[self.function]
        for module in modules:
            # Update module dependencies in imports
            self.section.imports.add(*module)

        calls = self.join_calls(arguments)

        if "__data['time']" in expression:
            # If the expression depens on time add to the dependencies
            merge_dependencies(calls, {"time": 1}, inplace=True)

        if "%(axis)s" in expression:
            # Vectorial expressions, compute the axis using dimensions
            # with ! operator
            if "%(1)s" in expression:
                subs = self.reorder(arguments)
                # NUMPY: following line may be avoided
                [arguments[i].reshape(self.section.subscripts, subs, True)
                 for i in ["0", "1"]]
            else:
                subs = arguments["0"].subscripts
            final_subscripts, arguments["axis"] = self._compute_axis(subs)

        elif "%(size)s" in expression:
            # Random expressions, need to give the final size of the
            # component to create one value per final coordinate
            final_subscripts = self.reorder(arguments, force="component")
            arguments["size"] = tuple(compute_shape(final_subscripts))
            if arguments["size"]:
                # Create an xarray from the random function output
                # NUMPY: not necessary
                # generate an xarray from the output
                subs = self.section.subscripts.simplify_subscript_input(
                    self.def_subs)[1]
                expression = f"xr.DataArray({expression}, {subs}, "\
                    f"{list(self.def_subs)})"
            calls["time"] = 1

        elif self.function == "active_initial":
            # Ee need to ensure that active initial outputs are always the
            # same and update dependencies as stateful object
            name = self.section.namespace.make_python_identifier(
                self.element.identifier, prefix="_active_initial")
            final_subscripts = self.reorder(arguments, force="equal")
            self.element.other_dependencies[name] = {
                "initial": arguments["1"].calls,
                "step": arguments["0"].calls
            }

            calls = {name: 1}

        elif self.function == "elmcount":
            final_subscripts = {}

        else:
            final_subscripts = self.reorder(arguments)
            if self.function == "xidz" and final_subscripts:
                # xidz must always return the same shape object
                if not arguments["1"].subscripts:
                    [arguments[i].reshape(
                        self.section.subscripts, final_subscripts, True)
                     for i in ["0", "1"]]
                elif arguments["0"].subscripts or arguments["2"].subscripts:
                    # NUMPY: not need this statement
                    [arguments[i].reshape(
                        self.section.subscripts, final_subscripts, True)
                     for i in ["0", "1", "2"]
                     if arguments[i].subscripts]
            elif self.function == "zidz" and final_subscripts:
                # zidz must always return the same shape object
                arguments["0"].reshape(
                    self.section.subscripts, final_subscripts, True)
                if arguments["1"].subscripts:
                    # NUMPY: not need this statement
                    arguments["1"].reshape(
                        self.section.subscripts, final_subscripts, True)
            elif self.function == "if_then_else" and final_subscripts:
                # if_then_else must always return the same shape object
                if not arguments["0"].subscripts:
                    # condition is a float
                    [arguments[i].reshape(
                        self.section.subscripts, final_subscripts, True)
                     for i in ["1", "2"]]
                else:
                    # condition has dimensions
                    [arguments[i].reshape(
                        self.section.subscripts, final_subscripts, True)
                     for i in ["0", "1", "2"]]

        return BuildAST(
            expression=expression % arguments,
            calls=calls,
            subscripts=final_subscripts,
            order=0)

    def _compute_axis(self, subscripts: dict) -> tuple:
        """
        Compute the axis to apply a vectorial function.

        Parameters
        ----------
        subscripts: dict
            The final_subscripts after reordering all the elements.

        Returns
        -------
        coords: dict
            The final coordinates after executing the vectorial function
        axis: list
            The list of dimensions to apply the function. Uses the
            dimensions with "!" at the end.

        """
        axis = []
        coords = {}
        for subs in subscripts:
            if subs.endswith("!"):
                # dimensions to apply along
                axis.append(subs)
            else:
                # dimensions remaining
                coords[subs] = subscripts[subs]
        return coords, axis


class AllocateAvailableBuilder(StructureBuilder):
    """Builder for allocate_available function."""

    def __init__(self, allocate_str: AllocateAvailableStructure,
                 component: object):
        super().__init__(None, component)

        pp = allocate_str.pp
        pp_sub = self.section.subscripts.elements[pp.reference][-1:]
        pp.subscripts.subscripts = pp.subscripts.subscripts[:-1] + pp_sub
        self.arguments = {
            "request": allocate_str.request,
            "pp": pp,
            "avail": allocate_str.avail
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.section.imports.add("allocation", "allocate_available")

        calls = self.join_calls(arguments)

        # the last sub of the request must be keep last sub of request and
        # priority
        last_sub = list(arguments["request"].subscripts)[-1]
        pp_sub = list(arguments["pp"].subscripts)[-1]
        # compute the merged subscripts
        final_subscripts = self.get_final_subscripts(arguments)
        # remove last sub from request
        last_sub_value = final_subscripts[last_sub]
        pp_sub_value = final_subscripts[pp_sub]
        del final_subscripts[last_sub], final_subscripts[pp_sub]

        # Update the susbcripts of avail
        arguments["avail"].reshape(
            self.section.subscripts, final_subscripts, True)

        # Include last sub of request in the last position and update
        # the subscripts of request
        final_subscripts[last_sub] = last_sub_value
        arguments["request"].reshape(
            self.section.subscripts, final_subscripts, True)

        # Include priority subscripts and update the subscripts of pp
        final_subscripts[pp_sub] = pp_sub_value
        arguments["pp"].reshape(
            self.section.subscripts, final_subscripts, True)

        expression = "allocate_available(%(request)s, %(pp)s, %(avail)s)"
        return BuildAST(
            expression=expression % arguments,
            calls=calls,
            subscripts=arguments["request"].subscripts,
            order=0)


class AllocateByPriorityBuilder(StructureBuilder):
    """Builder for allocate_by_priority function."""

    def __init__(self, allocate_str: AllocateByPriorityStructure,
                 component: object):
        super().__init__(None, component)
        self.arguments = {
            "request": allocate_str.request,
            "priority": allocate_str.priority,
            "width": allocate_str.width,
            "supply": allocate_str.supply
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.section.imports.add("allocation", "allocate_by_priority")

        calls = self.join_calls(arguments)

        # the last sub of the request must be keep last sub of request and
        # priority
        last_sub = list(arguments["request"].subscripts)[-1]
        # compute the merged subscripts
        final_subscripts = self.get_final_subscripts(arguments)
        # remove last sub from request
        last_sub_value = final_subscripts[last_sub]
        del final_subscripts[last_sub]

        # Update the susbcripts of width and supply
        arguments["width"].reshape(
            self.section.subscripts, final_subscripts, True)
        arguments["supply"].reshape(
            self.section.subscripts, final_subscripts, True)

        # Include last sub of request in the last position and update
        # the subscripts of request and priority
        final_subscripts[last_sub] = last_sub_value
        arguments["request"].reshape(
            self.section.subscripts, final_subscripts, True)
        arguments["priority"].reshape(
            self.section.subscripts, final_subscripts, True)

        expression = "allocate_by_priority(%(request)s, %(priority)s, "\
                     "%(width)s, %(supply)s)"
        return BuildAST(
            expression=expression % arguments,
            calls=calls,
            subscripts=arguments["request"].subscripts,
            order=0)


class ExtLookupBuilder(StructureBuilder):
    """Builder for External Lookups."""
    def __init__(self, getlookup_str: GetLookupsStructure,  component: object):
        super().__init__(None, component)
        self.file = getlookup_str.file
        self.tab = getlookup_str.tab
        self.x_row_or_col = getlookup_str.x_row_or_col
        self.cell = getlookup_str.cell
        self.arguments = {}

    def build(self, arguments: dict) -> Union[BuildAST, None]:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST or None
            The built object, unless the component has been added to an
            existing object using the 'add' method.

        """
        self.component.type = "Lookup"
        self.component.subtype = "External"
        arguments["params"] = "r'%s', '%s', '%s', '%s'" % (
            self.file, self.tab, self.x_row_or_col, self.cell
        )
        final_subs, arguments["subscripts"] =\
            self.section.subscripts.simplify_subscript_input(
                self.def_subs, self.element.subscripts)

        if "ext_lookups" in self.element.objects:
            # Object already exists, use 'add' method
            self.element.objects["ext_lookups"]["expression"] += "\n\n"\
                + self.element.objects["ext_lookups"]["name"]\
                + ".add(%(params)s, %(subscripts)s)" % arguments

            self.update_object_subscripts("ext_lookups", final_subs)

            return None
        else:
            # Create a new object
            self.section.imports.add("external", "ExtLookup")

            arguments["name"] = self.section.namespace.make_python_identifier(
                self.element.identifier, prefix="_ext_lookup")
            arguments["final_subs"] =\
                self.section.subscripts.simplify_subscript_input(
                    self.element.subs_dict)[1]
            self.component.subscripts_dict = [final_subs]

            self.element.objects["ext_lookups"] = {
                "name": arguments["name"],
                "expression": "%(name)s = ExtLookup(%(params)s, "
                              "%(subscripts)s, _root, "
                              "%(final_subs)s , '%(name)s')" % arguments,
                "component": self.component,
                "final_subs": final_subs
            }

            return BuildAST(
                expression=arguments["name"] + "(x, final_subs)",
                calls={
                    "__external__": arguments["name"],
                    "__lookup__": arguments["name"]
                },
                subscripts=final_subs,
                order=0)


class ExtDataBuilder(StructureBuilder):
    """Builder for External Data."""
    def __init__(self, getdata_str: GetDataStructure,  component: object):
        super().__init__(None, component)
        self.file = getdata_str.file
        self.tab = getdata_str.tab
        self.time_row_or_col = getdata_str.time_row_or_col
        self.cell = getdata_str.cell
        self.keyword = component.keyword
        self.arguments = {}

    def build(self, arguments: dict) -> Union[BuildAST, None]:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST or None
            The built object, unless the component has been added to an
            existing object using the 'add' method.

        """
        self.component.type = "Data"
        self.component.subtype = "External"
        arguments["params"] = "r'%s', '%s', '%s', '%s'" % (
            self.file, self.tab, self.time_row_or_col, self.cell
        )
        final_subs, arguments["subscripts"] =\
            self.section.subscripts.simplify_subscript_input(
                self.def_subs, self.element.subscripts)
        arguments["method"] = "'%s'" % self.keyword if self.keyword else None

        if "ext_data" in self.element.objects:
            # Object already exists, use add method
            self.element.objects["ext_data"]["expression"] += "\n\n"\
                + self.element.objects["ext_data"]["name"]\
                + ".add(%(params)s, %(method)s, %(subscripts)s)" % arguments

            self.update_object_subscripts("ext_data", final_subs)

            return None
        else:
            # Create a new object
            self.section.imports.add("external", "ExtData")

            arguments["name"] = self.section.namespace.make_python_identifier(
                self.element.identifier, prefix="_ext_data")
            arguments["final_subs"] =\
                self.section.subscripts.simplify_subscript_input(
                    self.element.subs_dict)[1]
            self.component.subscripts_dict = [final_subs]

            self.element.objects["ext_data"] = {
                "name": arguments["name"],
                "expression": "%(name)s = ExtData(%(params)s, "
                              " %(method)s, %(subscripts)s, "
                              "_root, %(final_subs)s ,'%(name)s')" % arguments,
                "component": self.component,
                "final_subs": final_subs
            }

            return BuildAST(
                expression=arguments["name"] + "(time())",
                calls={
                    "__external__": arguments["name"],
                    "__data__": arguments["name"],
                    "time": 1},
                subscripts=final_subs,
                order=0)


class ExtConstantBuilder(StructureBuilder):
    """Builder for External Constants."""
    def __init__(self, getconstant_str: GetConstantsStructure,
                 component: object):
        super().__init__(None, component)
        self.file = getconstant_str.file
        self.tab = getconstant_str.tab
        self.cell = getconstant_str.cell
        self.arguments = {}

    def build(self, arguments: dict) -> Union[BuildAST, None]:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST or None
            The built object, unless the component has been added to an
            existing object using the 'add' method.

        """
        self.component.type = "Constant"
        self.component.subtype = "External"
        arguments["params"] = "r'%s', '%s', '%s'" % (
            self.file, self.tab, self.cell
        )
        final_subs, arguments["subscripts"] =\
            self.section.subscripts.simplify_subscript_input(
                self.def_subs, self.element.subscripts)

        if "constants" in self.element.objects:
            # Object already exists, use 'add' method
            self.element.objects["constants"]["expression"] += "\n\n"\
                + self.element.objects["constants"]["name"]\
                + ".add(%(params)s, %(subscripts)s)" % arguments

            self.update_object_subscripts("constants", final_subs)

            return None
        else:
            # Create a new object
            self.section.imports.add("external", "ExtConstant")

            arguments["name"] = self.section.namespace.make_python_identifier(
                self.element.identifier, prefix="_ext_constant")
            arguments["final_subs"] =\
                self.section.subscripts.simplify_subscript_input(
                    self.element.subs_dict)[1]
            self.component.subscripts_dict = [final_subs]

            self.element.objects["constants"] = {
                "name": arguments["name"],
                "expression": "%(name)s = ExtConstant(%(params)s, "
                              "%(subscripts)s, _root, %(final_subs)s, "
                              "'%(name)s')" % arguments,
                "component": self.component,
                "final_subs": final_subs
            }

            return BuildAST(
                expression=arguments["name"] + "()",
                calls={"__external__": arguments["name"]},
                subscripts=final_subs,
                order=0)


class TabDataBuilder(StructureBuilder):
    """Builder for empty DATA expressions."""
    def __init__(self, data_str: DataStructure,  component: object):
        super().__init__(None, component)
        self.keyword = component.keyword
        self.arguments = {}

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.section.imports.add("data", "TabData")

        final_subs, arguments["subscripts"] =\
            self.section.subscripts.simplify_subscript_input(
                self.def_subs, self.element.subscripts)

        arguments["real_name"] = self.element.name
        arguments["py_name"] =\
            self.section.namespace.namespace[self.element.name]
        arguments["method"] = "'%s'" % self.keyword if self.keyword else None

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_data")

        # Create TabData object
        self.element.objects["tab_data"] = {
            "name": arguments["name"],
            "expression": "%(name)s = TabData('%(real_name)s', '%(py_name)s', "
                          "%(subscripts)s,  %(method)s)" % arguments
        }

        return BuildAST(
            expression=arguments["name"] + "(time())",
            calls={"time": 1, "__data__": arguments["name"]},
            subscripts=final_subs,
            order=0)


class InitialBuilder(StructureBuilder):
    """Builder for Initials."""
    def __init__(self, initial_str: InitialStructure, component: object):
        super().__init__(None, component)
        self.arguments = {
            "initial": initial_str.initial
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "Initial"
        self.section.imports.add("statefuls", "Initial")

        arguments["initial"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_initial")

        # Create the object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = Initial(lambda: %(initial)s, "
                          "'%(name)s')" % arguments,
        }
        # Add other-dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": arguments["initial"].calls,
            "step": {}
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class IntegBuilder(StructureBuilder):
    """Builder for Integs/Stocks."""
    def __init__(self, integ_str: IntegStructure, component: object):
        super().__init__(None, component)
        self.arguments = {
            "flow": integ_str.flow,
            "initial": integ_str.initial
        }
        self.non_negative = integ_str.non_negative

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "Integ"

        arguments["initial"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["flow"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_integ")

        # Create the object
        if self.non_negative:
            # Non-negative stocks
            self.section.imports.add("statefuls", "NonNegativeInteg")
            self.element.objects[arguments["name"]] = {
                "name": arguments["name"],
                "expression": "%(name)s = NonNegativeInteg("
                              "lambda: %(flow)s, "
                              "lambda: %(initial)s, '%(name)s')" % arguments
            }
        else:
            # Regular stocks
            self.section.imports.add("statefuls", "Integ")
            self.element.objects[arguments["name"]] = {
                "name": arguments["name"],
                "expression": "%(name)s = Integ(lambda: %(flow)s, "
                              "lambda: %(initial)s, '%(name)s')" % arguments
            }

        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": arguments["initial"].calls,
            "step": arguments["flow"].calls
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class DelayBuilder(StructureBuilder):
    """Builder for regular Delays."""
    def __init__(self, dtype: str,
                 delay_str: Union[DelayStructure, DelayNStructure],
                 component: object):
        super().__init__(None, component)
        self.arguments = {
            "input": delay_str.input,
            "delay_time": delay_str.delay_time,
            "initial": delay_str.initial,
            "order": delay_str.order
        }
        self.dtype = dtype

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "Delay"
        self.section.imports.add("statefuls", self.dtype)

        arguments["input"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["delay_time"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["initial"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix=f"_{self.dtype.lower()}")
        arguments["dtype"] = self.dtype

        # Add the object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = %(dtype)s(lambda: %(input)s, "
                          "lambda: %(delay_time)s, lambda: %(initial)s, "
                          "lambda: %(order)s, "
                          "time_step, '%(name)s')" % arguments,
        }
        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": merge_dependencies(
                arguments["initial"].calls,
                arguments["delay_time"].calls,
                arguments["order"].calls),
            "step": merge_dependencies(
                arguments["input"].calls,
                arguments["delay_time"].calls)

        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class DelayFixedBuilder(StructureBuilder):
    """Builder for Delay Fixed."""
    def __init__(self, delay_str: DelayFixedStructure, component: object):
        super().__init__(None, component)
        self.arguments = {
            "input": delay_str.input,
            "delay_time": delay_str.delay_time,
            "initial": delay_str.initial,
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "DelayFixed"
        self.section.imports.add("statefuls", "DelayFixed")

        arguments["input"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["initial"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_delayfixed")

        # Create object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = DelayFixed(lambda: %(input)s, "
                          "lambda: %(delay_time)s, lambda: %(initial)s, "
                          "time_step, '%(name)s')" % arguments,
        }
        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": merge_dependencies(
                arguments["initial"].calls,
                arguments["delay_time"].calls),
            "step": arguments["input"].calls
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class SmoothBuilder(StructureBuilder):
    """Builder for Smooths."""
    def __init__(self, smooth_str: Union[SmoothStructure, SmoothNStructure],
                 component: object):
        super().__init__(None, component)
        self.arguments = {
            "input": smooth_str.input,
            "smooth_time": smooth_str.smooth_time,
            "initial": smooth_str.initial,
            "order": smooth_str.order
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "Smooth"
        self.section.imports.add("statefuls", "Smooth")

        arguments["input"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["smooth_time"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["initial"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_smooth")

        # TODO in the future we need to ad timestep to show warnings about
        # the smooth time as its done with delays (see vensim help for smooth)
        # TODO in the future we may want to have 2 py_backend classes for
        # smooth as the behaviour is different for SMOOTH and SMOOTH N when
        # using RingeKutta scheme

        # Create object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = Smooth(lambda: %(input)s, "
                          "lambda: %(smooth_time)s, lambda: %(initial)s, "
                          "lambda: %(order)s, '%(name)s')" % arguments,
        }
        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": merge_dependencies(
                arguments["initial"].calls,
                arguments["order"].calls),
            "step": merge_dependencies(
                arguments["input"].calls,
                arguments["smooth_time"].calls)
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class TrendBuilder(StructureBuilder):
    """Builder for Trends."""
    def __init__(self, trend_str: TrendStructure, component: object):
        super().__init__(None, component)
        self.arguments = {
            "input": trend_str.input,
            "average_time": trend_str.average_time,
            "initial_trend": trend_str.initial_trend,
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "Trend"
        self.section.imports.add("statefuls", "Trend")

        arguments["input"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["average_time"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["initial_trend"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_trend")

        # Create object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = Trend(lambda: %(input)s, "
                          "lambda: %(average_time)s, "
                          "lambda: %(initial_trend)s, "
                          "'%(name)s')" % arguments,
        }
        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": merge_dependencies(
                arguments["initial_trend"].calls,
                arguments["input"].calls,
                arguments["average_time"].calls),
            "step": merge_dependencies(
                arguments["input"].calls,
                arguments["average_time"].calls)
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class ForecastBuilder(StructureBuilder):
    """Builder for Forecasts."""
    def __init__(self, forecast_str: ForecastStructure, component: object):
        super().__init__(None, component)
        self.arguments = {
            "input": forecast_str.input,
            "average_time": forecast_str.average_time,
            "horizon": forecast_str.horizon,
            "initial_trend": forecast_str.initial_trend
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "Forecast"
        self.section.imports.add("statefuls", "Forecast")

        arguments["input"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["average_time"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["horizon"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["initial_trend"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_forecast")

        # Create object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = Forecast(lambda: %(input)s, "
                          "lambda: %(average_time)s, lambda: %(horizon)s, "
                          "lambda: %(initial_trend)s, '%(name)s')" % arguments,
        }
        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial": merge_dependencies(
                arguments["input"].calls,
                arguments["initial_trend"].calls),
            "step": merge_dependencies(
                arguments["input"].calls,
                arguments["average_time"].calls,
                arguments["horizon"].calls)
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class SampleIfTrueBuilder(StructureBuilder):
    """Builder for Sample If True."""
    def __init__(self, sampleiftrue_str: SampleIfTrueStructure,
                 component: object):
        super().__init__(None, component)
        self.arguments = {
            "condition": sampleiftrue_str.condition,
            "input": sampleiftrue_str.input,
            "initial": sampleiftrue_str.initial,
        }

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Stateful"
        self.component.subtype = "SampleIfTrue"
        self.section.imports.add("statefuls", "SampleIfTrue")

        arguments["condition"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["input"].reshape(
            self.section.subscripts, self.def_subs, True)
        arguments["initial"].reshape(
            self.section.subscripts, self.def_subs, True)

        arguments["name"] = self.section.namespace.make_python_identifier(
            self.element.identifier, prefix="_sampleiftrue")

        # Create object
        self.element.objects[arguments["name"]] = {
            "name": arguments["name"],
            "expression": "%(name)s = SampleIfTrue(lambda: %(condition)s, "
                          "lambda: %(input)s, lambda: %(initial)s, "
                          "'%(name)s')" % arguments,
        }
        # Add other dependencies
        self.element.other_dependencies[arguments["name"]] = {
            "initial":
                arguments["initial"].calls,
            "step": merge_dependencies(
                arguments["condition"].calls,
                arguments["input"].calls)
        }

        return BuildAST(
            expression=arguments["name"] + "()",
            calls={arguments["name"]: 1},
            subscripts=self.def_subs,
            order=0)


class LookupsBuilder(StructureBuilder):
    """Builder for regular Lookups."""
    def __init__(self, lookups_str: LookupsStructure, component: object):
        super().__init__(None, component)
        self.arguments = {}
        self.x = lookups_str.x
        self.y = lookups_str.y
        self.keyword = lookups_str.type

    def build(self, arguments: dict) -> Union[BuildAST, None]:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST or None
            The built object, unless the component has been added to an
            existing object using the 'add' method.

        """
        self.component.type = "Lookup"
        self.component.subtype = "Normal"
        # Get the numeric values as numpy arrays
        arguments["x"] = np.array2string(
            np.array(self.x),
            separator=",",
            threshold=len(self.x)
        )
        arguments["y"] = np.array2string(
            np.array(self.y),
            separator=",",
            threshold=len(self.y)
        )
        arguments["subscripts"] = self.def_subs
        arguments["interp"] = self.keyword

        if "hardcoded_lookups" in self.element.objects:
            # Object already exists, use 'add' method
            self.element.objects["hardcoded_lookups"]["expression"] += "\n\n"\
                + self.element.objects["hardcoded_lookups"]["name"]\
                + ".add(%(x)s, %(y)s, %(subscripts)s)" % arguments

            return None
        else:
            # Create a new object
            self.section.imports.add("lookups", "HardcodedLookups")

            arguments["name"] = self.section.namespace.make_python_identifier(
                self.element.identifier, prefix="_hardcodedlookup")
            arguments["final_subs"] =\
                self.section.subscripts.simplify_subscript_input(
                    self.element.subs_dict)[1]

            self.element.objects["hardcoded_lookups"] = {
                "name": arguments["name"],
                "expression": "%(name)s = HardcodedLookups(%(x)s, %(y)s, "
                              "%(subscripts)s, '%(interp)s', "
                              "%(final_subs)s, '%(name)s')"
                              % arguments,
                "final_subs": self.element.subs_dict
            }

            return BuildAST(
                expression=arguments["name"] + "(x, final_subs)",
                calls={"__lookup__": arguments["name"]},
                subscripts=self.def_subs,
                order=0)


class InlineLookupsBuilder(StructureBuilder):
    """Builder for inline Lookups."""
    def __init__(self, inlinelookups_str: InlineLookupsStructure,
                 component: object):
        super().__init__(None, component)
        self.arguments = {
            "value": inlinelookups_str.argument
        }
        self.lookups = inlinelookups_str.lookups

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.component.type = "Auxiliary"
        self.component.subtype = "with Lookup"
        self.section.imports.add("numpy")
        # Get the numeric values as numpy arrays
        arguments["x"] = np.array2string(
            np.array(self.lookups.x),
            separator=",",
            threshold=len(self.lookups.x)
        )
        arguments["y"] = np.array2string(
            np.array(self.lookups.y),
            separator=",",
            threshold=len(self.lookups.y)
        )
        if arguments["value"].subscripts:
            subs = arguments["value"].subscripts
            expression = "np.interp(%(value)s, %(x)s, %(y)s)" % arguments
            return BuildAST(
                expression="xr.DataArray(%s, %s, %s)" % (
                    expression, subs, list(subs)),
                calls=arguments["value"].calls,
                subscripts=subs,
                order=0)
        else:
            return BuildAST(
                expression="np.interp(%(value)s, %(x)s, %(y)s)" % arguments,
                calls=arguments["value"].calls,
                subscripts={},
                order=0)


class ReferenceBuilder(StructureBuilder):
    """Builder for references to other variables."""
    def __init__(self, reference_str: ReferenceStructure, component: object):
        super().__init__(None, component)
        self.mapping_subscripts = {}
        self.reference = reference_str.reference
        self.subscripts = reference_str.subscripts
        self.arguments = {}

    @property
    def subscripts(self):
        return self._subscripts

    @subscripts.setter
    def subscripts(self, subscripts: SubscriptsReferenceStructure):
        """Get subscript dictionary from reference"""
        ref_subs = getattr(subscripts, "subscripts", [])

        self._subscripts = self.section.subscripts.make_coord_dict(ref_subs)

        if len(ref_subs) != len(self._subscripts):
            # The reference has repeated subscript ranges, this is
            # not compatible with Python as we use dictionaries to save
            # the subscript ranges information (duplicates a key)

            # Get the original name of the reference
            origin_ref = self.reference
            for origin_ref2 in self.section.namespace.namespace.keys():
                if origin_ref == origin_ref2.lower().replace(" ", "_"):
                    origin_ref = origin_ref2
                    break

            warnings.warn(
                "The reference to '%s' in variable '%s' has duplicated "
                "subscript ranges. If mapping is used in one of them, "
                "please, rewrite reference subscripts to avoid "
                "duplicates. Otherwise, the final model may crash..."
                % (origin_ref, self.element.name)
            )

        # get the subscripts after applying the mapping if necessary
        for dim, coordinates in self._subscripts.items():
            if len(coordinates) > 1:
                # we create the mapping only with those subscripts that are
                # ranges as we need to ignore singular subscripts because
                # that dimension is removed from final element
                if dim not in self.def_subs and not dim.endswith("!"):
                    # the reference has a subscripts which is it not
                    # applied (!) and does not appear in the definition
                    # of the variable
                    not_mapped = True
                    for mapped in self.section.subscripts.mapping[dim]:
                        # check the mapped subscripts
                        # TODO update this and the parser to make it
                        # compatible with more complex mappings
                        if mapped in self.def_subs\
                          and mapped not in self._subscripts:
                            # the mapped subscript appears in the definition
                            # and it is not already in the variable
                            self.mapping_subscripts[mapped] =\
                                self.section.subscripts.subscripts[mapped]
                            not_mapped = False
                            break
                    if not_mapped:
                        # manage other not mapped subscripts
                        # this is necessary for Allocate Available
                        # where we must force the expression in the
                        # right to have more subscripts thant the
                        # expression in the left
                        self.mapping_subscripts[dim] = coordinates
                else:
                    # the subscript is in the variable definition,
                    # do not change it
                    self.mapping_subscripts[dim] = coordinates

    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        if self.reference not in self.section.namespace.cleanspace:
            # Manage references to subscripts (subscripts used as variables)
            expression, subscripts =\
                self.section.subscripts.subscript2num[self.reference]
            subscripts_out = self.section.subscripts.simplify_subscript_input(
                subscripts)[1]
            if subscripts:
                self.section.imports.add("numpy")
                # NUMPY: not need this if
                expression = "xr.DataArray(%s, %s, %s)" % (
                    expression, subscripts_out, list(subscripts))
            return BuildAST(
                expression=expression,
                calls={},
                subscripts=subscripts,
                order=0)

        reference = self.section.namespace.cleanspace[self.reference]

        expression = reference + "()"

        if not self.subscripts:
            return BuildAST(
                expression=expression,
                calls={reference: 1},
                subscripts={},
                order=0)

        original_subs = self.section.subscripts.make_coord_dict(
                    self.section.subscripts.elements[reference])

        expression, final_subs = self._visit_subscripts(
            expression, original_subs)

        return BuildAST(
            expression=expression,
            calls={reference: 1},
            subscripts=final_subs,
            order=0)

    def _visit_subscripts(self, expression: str, original_subs: dict) -> tuple:
        """
        Visit the subcripts of a reference to subset a subarray if neccessary
        or apply mapping.

        Parameters
        ----------
        expression: str
            The expression of visiting the variable.
        original_subs: dict
            The original subscript dict of the variable.

        Returns
        -------
        expression: str
            The expression with the necessary operations.
        mapping_subscirpts: dict
            The final subscripts of the reference after applying mapping.

        """
        loc, rename, final_subs, reset_coords, to_float =\
            visit_loc(self.subscripts, original_subs)

        if loc is not None:
            # NUMPY: expression += "[%s]" % ", ".join(loc)
            expression += f".loc[{loc}]"
        if to_float:
            # NUMPY: Not neccessary
            expression = "float(" + expression + ")"
        elif reset_coords:
            # NUMPY: Not neccessary
            expression += ".reset_coords(drop=True)"
        if rename:
            # NUMPY: Not neccessary
            expression += ".rename(%s)" % rename

        # NUMPY: This will not be necessary, we only need to return
        #        self.mapping_subscripts
        if self.mapping_subscripts != final_subs:
            subscripts_out = self.section.subscripts.simplify_subscript_input(
                self.mapping_subscripts)[1]
            expression = "xr.DataArray(%s.values, %s, %s)" % (
                expression, subscripts_out, list(self.mapping_subscripts)
            )

        return expression, self.mapping_subscripts


class NumericBuilder(StructureBuilder):
    """Builder for numeric and nan values."""
    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        if np.isnan(self.value):
            self.section.imports.add("numpy")

            return BuildAST(
                expression="np.nan",
                calls={},
                subscripts={},
                order=0)
        else:
            return BuildAST(
                expression=repr(self.value),
                calls={},
                subscripts={},
                order=0)


class ArrayBuilder(StructureBuilder):
    """Builder for arrays."""
    def build(self, arguments: dict) -> BuildAST:
        """
        Build method.

        Parameters
        ----------
        arguments: dict
            The dictionary of builded arguments.

        Returns
        -------
        built_ast: BuildAST
            The built object.

        """
        self.value = np.array2string(
            self.value.reshape(compute_shape(self.def_subs)),
            separator=",",
            threshold=np.prod(self.value.shape)
        )
        self.component.type = "Constant"
        self.component.subtype = self.component.subtype or "Normal"

        final_subs, subscripts_out =\
            self.section.subscripts.simplify_subscript_input(
                self.def_subs, self.element.subscripts)

        return BuildAST(
            expression="xr.DataArray(%s, %s, %s)" % (
                self.value, subscripts_out, list(final_subs)),
            calls={},
            subscripts=final_subs,
            order=0)


def merge_dependencies(*dependencies: dict, inplace: bool = False) -> dict:
    """
    Merge two dependencies dicts of an element.

    Parameters
    ----------
    dependencies: dict
        The dictionaries of dependencies to merge.

    inplace: bool (optional)
        If True the final dependencies dict will be updated in the first
        dependencies argument, mutating it. Default is False.

    Returns
    -------
    current: dict
        The final dependencies dict.

    """
    current = dependencies[0]
    if inplace:
        current = dependencies[0]
    else:
        current = dependencies[0].copy()
    for new in dependencies[1:]:
        if not current:
            current.update(new)
        elif new:
            # regular element
            current_set, new_set = set(current), set(new)
            for dep in current_set.intersection(new_set):
                # if dependency is in both sum the number of calls
                if dep.startswith("__"):
                    # if it is special (__lookup__, __external__) continue
                    continue
                else:
                    current[dep] += new[dep]
            for dep in new_set.difference(current_set):
                # if dependency is only in new copy it
                current[dep] = new[dep]

    return current


def visit_loc(current_subs: dict, original_subs: dict,
              keep_shape: bool = False) -> tuple:
    """
    Compares the original subscripts and the current subscripts and
    returns subindexing information if needed.

    Parameters
    ----------
    current_subs: dict
        The dictionary of the subscripts that are used in the variable.

    original_subs: dict
        The dictionary of the original subscripts of the variable.

    keep_shape: bool (optional)
        If True will keep the number of dimensions of the original element
        and return only loc. Default is False.

    Returns
    -------
    loc: list of str or None
        List of the subscripting in each dimensions. If all are full (":"),
        None is rerned wich means that array indexing is not needed.

    rename: dict
        Dictionary of the dimensions to rename.

    final_subs: dict
        Dictionary of the final subscripts of the variable.

    reset_coords: bool
        Boolean indicating if the coords need to be reseted.

    to_float: bool
        Boolean indicating if the variable should be converted to a float.

    """
    final_subs, rename, loc, reset_coords, to_float = {}, {}, [], False, True
    subscripts_zipped = zip(current_subs.items(), original_subs.items())
    for (dim, coord), (orig_dim, orig_coord) in subscripts_zipped:
        if len(coord) == 1:
            # subset a 1 dimension value
            # NUMPY: subset value [:, N, :, :]
            if keep_shape:
                # NUMPY: not necessary
                loc.append(f"[{repr(coord[0])}]")
            else:
                loc.append(repr(coord[0]))
            reset_coords = True
        elif len(coord) < len(orig_coord):
            # subset a subrange
            # NUMPY: subset value [:, :, np.array([1, 0]), :]
            # NUMPY: as order may change we need to check if
            #        dim != orig_dim
            # NUMPY: use also ranges [:, :, 2:5, :] when possible
            if dim.endswith("!"):
                loc.append("_subscript_dict['%s']" % dim[:-1])
            else:
                if dim != orig_dim:
                    loc.append("_subscript_dict['%s']" % dim)
                else:
                    # workaround for locs from external objects merge
                    loc.append(repr(coord))
            final_subs[dim] = coord
            to_float = False
        else:
            # do nothing
            # NUMPY: same, we can remove float = False
            loc.append(":")
            final_subs[dim] = coord
            to_float = False

        if dim != orig_dim and len(coord) != 1:
            # NUMPY: check order of dimensions, make all subranges work
            #        with the same dimensions?
            # NUMPY: this could be solved in the previous if/then/else
            rename[orig_dim] = dim

    if all(dim == ":" for dim in loc):
        # if all are ":" then no need to loc
        loc = None
    else:
        loc = ", ".join(loc)

    if keep_shape:
        return loc

    # convert to float if also coords are reseted (input is an array)
    to_float = to_float and reset_coords

    # NUMPY: save and return only loc, the other are not needed
    return loc, rename, final_subs, reset_coords, to_float


class ASTVisitor:
    """
    ASTVisitor allows visiting the Abstract Synatx Tree of a component
    returning the Python object and generating the neccessary objects.

    Parameters
    ----------
    component: ComponentBuilder
        The component builder to build.

    """
    _builders = {
        InitialStructure: InitialBuilder,
        IntegStructure: IntegBuilder,
        DelayStructure: lambda x, y: DelayBuilder("Delay", x, y),
        DelayNStructure: lambda x, y: DelayBuilder("DelayN", x, y),
        DelayFixedStructure: DelayFixedBuilder,
        SmoothStructure: SmoothBuilder,
        SmoothNStructure: SmoothBuilder,
        TrendStructure: TrendBuilder,
        ForecastStructure: ForecastBuilder,
        SampleIfTrueStructure: SampleIfTrueBuilder,
        GetConstantsStructure: ExtConstantBuilder,
        GetDataStructure: ExtDataBuilder,
        GetLookupsStructure: ExtLookupBuilder,
        LookupsStructure: LookupsBuilder,
        InlineLookupsStructure: InlineLookupsBuilder,
        DataStructure: TabDataBuilder,
        ReferenceStructure: ReferenceBuilder,
        CallStructure: CallBuilder,
        GameStructure: GameBuilder,
        AllocateAvailableStructure: AllocateAvailableBuilder,
        AllocateByPriorityStructure: AllocateByPriorityBuilder,
        LogicStructure: OperationBuilder,
        ArithmeticStructure: OperationBuilder,
        int: NumericBuilder,
        float: NumericBuilder,
        np.ndarray: ArrayBuilder,
    }

    def __init__(self, component: object):
        # component typing should be ComponentBuilder, but importing it
        # for typing would create a circular dependency :S
        self.ast = component.ast
        self.subscripts = component.subscripts_dict
        self.component = component

    def visit(self) -> Union[None, BuildAST]:
        """
        Visit the Abstract Syntax Tree of the component.

        Returns
        -------
        visit_out: BuildAST or None
            The BuildAST object resulting from visiting the AST. If the
            component content has been added to an existing object
            using the 'add' method it will return None.

        """
        visit_out = self._visit(self.ast)

        if not visit_out:
            # external objects that are declared with other expression
            return None

        if not visit_out.calls and self.component.type == "Auxiliary":
            self.component.type = "Constant"
            self.component.subtype = "Normal"

        # include dependencies of the current component in the element
        merge_dependencies(
            self.component.element.dependencies,
            visit_out.calls,
            inplace=True)

        if not visit_out.subscripts:
            # expression is a float
            return visit_out

        # NUMPY not needed
        # get subscript in elements as name of the ranges may change
        subscripts_in_element = {
            dim: coords
            for dim, coords
            in zip(self.component.element.subscripts, self.subscripts.values())
        }

        reshape = (
            (visit_out.subscripts != self.subscripts
             or list(visit_out.subscripts) != list(self.subscripts))
            and
            (visit_out.subscripts != subscripts_in_element
             or list(visit_out.subscripts) != list(subscripts_in_element))
        )

        if reshape:
            # NUMPY: in this case we need to tile along dims if neccessary
            #        or reorder the dimensions
            visit_out.reshape(
                self.component.section.subscripts, self.subscripts, True)

        return visit_out

    def _visit(self, ast_object: AbstractSyntax) -> AbstractSyntax:
        """
        Visit one Builder and its arguments.
        """
        builder = self._builders[type(ast_object)](ast_object, self.component)
        arguments = {
            name: self._visit(value)
            for name, value in builder.arguments.items()
        }
        return builder.build(arguments)



"""
Benchmarking tools for testing and comparing outputs between different files.
Some of these functions are also used for testing.
"""
import warnings
from pathlib import Path

import numpy as np
import pandas as pd

from pysd import read_vensim, read_xmile, load
from ..py_backend.utils import load_outputs, detect_encoding
from pysd.translators.vensim.vensim_utils import supported_extensions as\
    vensim_extensions
from pysd.translators.xmile.xmile_utils import supported_extensions as\
    xmile_extensions


def runner(model_file, canonical_file=None, transpose=False, data_files=None):
    """
    Translates and runs a model and returns its output and the
    canonical output.

    Parameters
    ----------
    model_file: str
        Name of the original model file. Must be '.mdl' or '.xmile'.

    canonical_file: str or None (optional)
        Canonical output file to read. If None, will search for 'output.csv'
        and 'output.tab' in the model directory. Default is None.

    transpose: bool (optional)
        If True reads transposed canonical file, i.e. one variable per row.
        Default is False.

    data_files: list (optional)
        List of the data files needed to run the model.

    Returns
    -------
    output, canon: (pandas.DataFrame, pandas.DataFrame)
        pandas.DataFrame of the model output and the canonical output.

    """
    if isinstance(model_file, str):
        model_file = Path(model_file)

    directory = model_file.parent

    # load canonical output
    if not canonical_file:
        if directory.joinpath('output.csv').is_file():
            canonical_file = directory.joinpath('output.csv')
        elif directory.joinpath('output.tab').is_file():
            canonical_file = directory.joinpath('output.tab')
        else:
            raise FileNotFoundError("\nCanonical output file not found.")

    canon = load_outputs(canonical_file,
                         transpose=transpose,
                         encoding=detect_encoding(canonical_file))

    # load model
    if model_file.suffix.lower() in vensim_extensions:
        model = read_vensim(model_file, data_files)
    elif model_file.suffix.lower() in xmile_extensions:
        model = read_xmile(model_file, data_files)
    elif model_file.suffix.lower() == ".py":
        model = load(model_file, data_files)
    else:
        raise ValueError(
            "\nThe model file name must be a Vensim"
            f" ({', '.join(vensim_extensions)}), a Xmile "
            f"({', '.join(xmile_extensions)}) or a PySD (.py) model file...")

    # run model and return the result

    return model.run(return_columns=canon.columns), canon


def assert_frames_close(actual, expected, assertion="raise",
                        verbose=False, precision=2, **kwargs):
    """
    Compare DataFrame items by column and
    raise AssertionError if any column is not equal.

    Ordering of columns is unimportant, items are compared only by label.
    NaN and infinite values are supported.

    Parameters
    ----------
    actual: pandas.DataFrame
        Actual value from the model output.

    expected: pandas.DataFrame
        Expected model output.

    assertion: str (optional)
        "raise" if an error should be raised when not able to assert
        that two frames are close. If "warning", it will show a warning
        message. If "return" it will return information. Default is "raise".

    verbose: bool (optional)
        If True, if any column is not close the actual and expected values
        will be printed in the error/warning message with the difference.
        Default is False.

    precision: int (optional)
        Precision to print the numerical values of assertion verbosed message.
        Default is 2.

    kwargs:
        Optional rtol and atol values for assert_allclose.

    Returns
    -------
    (cols, first_false_time, first_false_cols) or None: (set, float, set) or None
        If assertion is 'return', return the sets of the all columns that are
        different. The time when the first difference was found and the
        variables that what different at that time. If assertion is not
        'return' it returns None.

    Examples
    --------
    >>> assert_frames_close(
    ...     pd.DataFrame(100, index=range(5), columns=range(3)),
    ...     pd.DataFrame(100, index=range(5), columns=range(3)))

    >>> assert_frames_close(
    ...     pd.DataFrame(100, index=range(5), columns=range(3)),
    ...     pd.DataFrame(110, index=range(5), columns=range(3)),
    ...     rtol=.2)

    >>> assert_frames_close(
    ...     pd.DataFrame(100, index=range(5), columns=range(3)),
    ...     pd.DataFrame(150, index=range(5), columns=range(3)),
    ...     rtol=.2)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    AssertionError:
    Following columns are not close:
    \t'0'

    >>> assert_frames_close(
    ...     pd.DataFrame(100, index=range(5), columns=range(3)),
    ...     pd.DataFrame(150, index=range(5), columns=range(3)),
    ...     verbose=True, rtol=.2)  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
    ...
    AssertionError:
    Following columns are not close:
    \t'0'
    Column '0' is not close.
    Expected values:
    \t[150, 150, 150, 150, 150]
    Actual values:
    \t[100, 100, 100, 100, 100]
    Difference:
    \t[50, 50, 50, 50, 50]

    >>> assert_frames_close(
    ...     pd.DataFrame(100, index=range(5), columns=range(3)),
    ...     pd.DataFrame(150, index=range(5), columns=range(3)),
    ...     rtol=.2, assertion="warn")
    ...
    UserWarning:
    Following columns are not close:
    \t'0'

    References
    ----------
    Derived from:
        http://nbviewer.jupyter.org/gist/jiffyclub/ac2e7506428d5e1d587b

    """
    if not isinstance(actual, pd.DataFrame)\
       or not isinstance(expected, pd.DataFrame):
        raise TypeError('\nInputs must both be pandas DataFrames.')

    expected_cols, actual_cols = set(expected.columns), set(actual.columns)

    if expected_cols != actual_cols:
        # columns are not equal
        message = ""

        if actual_cols.difference(expected_cols):
            columns = sorted([
                "'" + col + "'" for col
                in actual_cols.difference(expected_cols)])
            columns = ", ".join(columns)
            message += '\nColumns ' + columns\
                       + ' from actual values not found in expected values.'

        if expected_cols.difference(actual_cols):
            columns = sorted([
                "'" + col + "'" for col
                in expected_cols.difference(actual_cols)])
            columns = ", ".join(columns)
            message += '\nColumns ' + columns\
                       + ' from expected values not found in actual values.'

        if assertion == "raise":
            raise ValueError(
                '\nColumns from actual and expected values must be equal.'
                + message)
        else:
            warnings.warn(message)

    columns = list(actual_cols.intersection(expected_cols))

    # TODO let compare dataframes with different timestamps if "warn"
    assert np.all(np.equal(expected.index.values, actual.index.values)), \
        "test set and actual set must share a common index, "\
        "instead found %s vs %s" % (expected.index.values, actual.index.values)

    # if for Vensim outputs where constant values are only in the first row
    _remove_constant_nan(expected)
    _remove_constant_nan(actual)

    c = assert_allclose(expected[columns],
                        actual[columns],
                        **kwargs)

    if c.all().all():
        return (set(), np.nan, set()) if assertion == "return" else None

    # Get the columns that have the first different value, useful for
    # debugging
    false_index = c.apply(
        lambda x: np.where(~x)[0][0] if not x.all() else np.nan)
    index_first_false = int(np.nanmin(false_index))
    time_first_false = c.index[index_first_false]
    variable_first_false = sorted(
        false_index.index[false_index == index_first_false])

    columns = sorted(np.array(columns, dtype=str)[~c.all().values])

    assertion_details = "\nFollowing columns are not close:\n\t"\
                        + ", ".join(columns) + "\n\n"\
                        + f"First false values ({time_first_false}):\n\t"\
                        + ", ".join(variable_first_false)

    if verbose:
        for col in columns:
            assertion_details += '\n\n'\
                + f"Column '{col}' is not close."\
                + '\n\nExpected values:\n\t'\
                + np.array2string(expected[col].values,
                                  precision=precision,
                                  separator=', ')\
                + '\n\nActual values:\n\t'\
                + np.array2string(actual[col].values,
                                  precision=precision,
                                  separator=', ')\
                + '\n\nDifference:\n\t'\
                + np.array2string(expected[col].values-actual[col].values,
                                  precision=precision,
                                  separator=', ')

    if assertion == "raise":
        raise AssertionError(assertion_details)
    elif assertion == "return":
        return (set(columns), time_first_false, set(variable_first_false))
    else:
        warnings.warn(assertion_details)


def assert_allclose(x, y, rtol=1.e-5, atol=1.e-5):
    """
    Asserts if numeric values from two arrays are close.

    Parameters
    ----------
    x: ndarray
        Expected value.
    y: ndarray
        Actual value.
    rtol: float (optional)
        Relative tolerance on the error. Default is 1.e-5.
    atol: float (optional)
        Absolut tolerance on the error. Default is 1.e-5.

    Returns
    -------
    None

    """
    return ((abs(x - y) <= atol + rtol * abs(y)) + x.isna()*y.isna())


def _remove_constant_nan(df):
    """
    Removes nana values in constant value columns produced by Vensim
    """
    nan_cols = np.isnan(df.iloc[1:, :]).all().values
    df.loc[:, nan_cols] = df.loc[:, nan_cols].iloc[0].values



import warnings
import re
import random
from pathlib import Path

import numpy as np
import xarray as xr
import pandas as pd

from .utils import load_outputs


class Columns():
    """
    Class to save the read columns in data files
    """
    _files = {}

    @classmethod
    def read(cls, file_name, encoding=None):
        """
        Read the columns from the data file or return the previously read ones
        """
        file_name = Path(file_name)
        if file_name in cls._files:
            return cls._files[file_name]
        else:
            columns = cls.read_file(file_name, encoding)
            cls._files[file_name] = columns
            return columns

    @classmethod
    def read_file(cls, file_name, encoding=None):
        """
        Get the columns from an output csv or tab file.

        Parameters
        ----------
        file_name: str
            Output file to read. Must be csv or tab.

        encoding: str or None (optional)
            Encoding type to read output file. Needed if the file has special
            characters. Default is None.

        Returns
        -------
        out, transposed: list, bool
            The list of the columns in the output file and a boolean flag
            to indicate if the output file is transposed.

        """
        # in the most cases variables will be split per columns, then
        # read the first row to have all the column names
        out = cls.read_line(file_name, encoding)
        if out is None:
            raise ValueError(
                f"\nNot able to read '{str(file_name)}'. "
                + "Only '.csv', '.tab' files are accepted.")

        transpose = False

        try:
            # if we fail converting columns to float then they are
            # not numeric values, so current direction is okay
            [float(col) for col in random.sample(out, min(3, len(out)))]
            # we did not fail, read the first column to see if variables
            # are split per rows
            out = cls.read_col(file_name, encoding)
            transpose = True
            # if we still are able to transform values to float the
            # file is not valid
            [float(col) for col in random.sample(out, min(3, len(out)))]
        except ValueError:
            return out, transpose
        else:
            raise ValueError(
                f"Invalid file format '{str(file_name)}'... varible names "
                "should appear in the first row or in the first column...")

    @classmethod
    def read_line(cls, file_name, encoding=None):
        """
        Read the firts row and return a set of it.
        """
        if file_name.suffix.lower() == ".tab":
            return list(pd.read_table(file_name,
                                      nrows=0,
                                      encoding=encoding,
                                      dtype=str,
                                      header=0).iloc[:, 1:])
        elif file_name.suffix.lower() == ".csv":
            return list(pd.read_csv(file_name,
                                    nrows=0,
                                    encoding=encoding,
                                    dtype=str,
                                    header=0).iloc[:, 1:])
        else:
            return None

    @classmethod
    def read_col(cls, file_name, encoding=None):
        """
        Read the firts column and return a it.
        """
        if file_name.suffix.lower() == ".tab":
            return list(pd.read_table(file_name,
                                      usecols=[0],
                                      encoding=encoding,
                                      dtype=str).iloc[:, 0].to_list())
        elif file_name.suffix.lower() == ".csv":
            return list(pd.read_csv(file_name,
                                    usecols=[0],
                                    encoding=encoding,
                                    dtype=str).iloc[:, 0].to_list())

    @classmethod
    def get_columns(cls, file_name, vars=None, encoding=None):
        """
        Get columns names from a tab or csv file and return those that
        match with the given ones.

        Parameters
        ----------
        file_name: str
            Output file to read. Must be csv or tab.

        vars: list
            List of var names to find in the file.

        encoding: str or None (optional)
            Encoding type to read output file. Needed if the file has special
            characters. Default is None.

        Return
        ------
        columns, transpose: set, bool
            The set of columns as they are named in the input file and a
            boolean flag to indicate if the input file is transposed or
            not.

        """
        if vars is None:
            # Not var specified, return all available variables
            return cls.read(file_name, encoding)

        columns, transpose = cls.read(file_name, encoding)

        vars_extended = []
        for var in vars:
            vars_extended.append(var)
            if var.startswith('"') and var.endswith('"'):
                # the variables in "" are reded without " by pandas
                vars_extended.append(var[1:-1])

        outs = set()
        for var in columns:
            if var in vars_extended:
                # var is in vars_extended (no subscripts)
                outs.add(var)
                vars_extended.remove(var)
            else:
                for var1 in vars_extended:
                    if var.startswith(var1 + "["):
                        # var is subscripted
                        outs.add(var)

        return outs, transpose

    @classmethod
    def clean(cls):
        """
        Clean the dictionary of read files
        """
        cls._files = {}


class Data(object):
    # TODO add __init__ and use this class for used input pandas.Series
    # as Data
    # def __init__(self, data, coords, interp="interpolate"):

    def set_values(self, values):
        """Set new values from user input"""
        self.data = xr.DataArray(
            np.nan, self.final_coords, list(self.final_coords))

        if isinstance(values, pd.Series):
            index = list(values.index)
            index.sort()
            self.data = self.data.expand_dims(
                {'time': index}, axis=0).copy()

            for index, value in values.items():
                if isinstance(values.values[0], xr.DataArray):
                    self.data.loc[index].loc[value.coords] = value
                else:
                    self.data.loc[index] = value
        else:
            if isinstance(values, xr.DataArray):
                self.data.loc[values.coords] = values.values
            else:
                if self.final_coords:
                    self.data.loc[:] = values
                else:
                    self.data = values

    def __call__(self, time):
        try:
            if time in self.data['time'].values:
                outdata = self.data.sel(time=time)
            elif self.interp == "raw":
                return self.nan
            elif time > self.data['time'].values[-1]:
                warnings.warn(
                    self.py_name + "\n"
                    + "extrapolating data above the maximum value of the time")
                outdata = self.data[-1]
            elif time < self.data['time'].values[0]:
                warnings.warn(
                    self.py_name + "\n"
                    + "extrapolating data below the minimum value of the time")
                outdata = self.data[0]
            elif self.interp == "interpolate":
                outdata = self.data.interp(time=time)
            elif self.interp == 'look_forward':
                outdata = self.data.sel(time=time, method="backfill")
            elif self.interp == 'hold_backward':
                outdata = self.data.sel(time=time, method="pad")

            if self.is_float:
                # if data has no-coords return a float
                return float(outdata)
            else:
                # Remove time coord from the DataArray
                return outdata.reset_coords('time', drop=True)
        except (TypeError, KeyError):
            if self.data is None:
                raise ValueError(
                    self.py_name + "\n"
                    "Trying to interpolate data variable before loading"
                    " the data...")

            # this except catch the errors when a data has been
            # changed to a constant value by the user
            return self.data
        except Exception as err:
            raise err


class TabData(Data):
    """
    Data from tabular file tab/csv, it could be from Vensim output.
    """
    def __init__(self, real_name, py_name, coords, interp="interpolate"):
        self.real_name = real_name
        self.py_name = py_name
        self.coords = coords
        self.final_coords = coords
        self.interp = interp.replace(" ", "_") if interp else None
        self.is_float = not bool(coords)
        self.data = None

        if self.interp not in ["interpolate", "raw",
                               "look_forward", "hold_backward"]:
            raise ValueError(self.py_name + "\n"
                             + "The interpolation method (interp) must be "
                             + "'raw', 'interpolate', "
                             + "'look_forward' or 'hold_backward'")

    def load_data(self, file_names):
        """
        Load data values from files.

        Parameters
        ----------
        file_names: list or str or pathlib.Path
            Name of the files to search the variable in.

        Returns
        -------
        out: xarray.DataArray
            Resulting data array with the time in the first dimension.

        """
        if isinstance(file_names, (str, Path)):
            file_names = [file_names]

        for file_name in file_names:
            self.data = self._load_data(Path(file_name))
            if self.data is not None:
                break

        if self.data is None:
            raise ValueError(
                f"_data_{self.py_name}\n"
                f"Data for {self.real_name} not found in "
                f"{', '.join([str(file_name) for file_name in file_names])}")

    def _load_data(self, file_name):
        """
        Load data values from output

        Parameters
        ----------
        file_name: pathlib.Path
            Name of the file to search the variable in.

        Returns
        -------
        out: xarray.DataArray or None
            Resulting data array with the time in the first dimension.

        """
        # TODO inlcude missing values managment as External objects
        # get columns to load variable
        if file_name.suffix in [".csv", ".tab"]:

            columns, transpose = Columns.get_columns(
                file_name, vars=[self.real_name, self.py_name])

            if not columns:
                # the variable is not in the passed file
                return None

            if not self.coords:
                # 0 dimensional data
                self.nan = np.nan
                values = load_outputs(file_name, transpose, columns=columns)
                return xr.DataArray(
                    values.iloc[:, 0].values,
                    {'time': values.index.values},
                    ['time'])

            # subscripted data
            dims = list(self.coords)

            values = load_outputs(file_name, transpose, columns=columns)

            self.nan = xr.DataArray(np.nan, self.coords, dims)
            out = xr.DataArray(
                np.nan,
                {'time': values.index.values, **self.coords},
                ['time'] + dims)

            for column in values.columns:
                coords = {
                    dim: [coord]
                    for (dim, coord)
                    in zip(dims, re.split(r'\[|\]|\s*,\s*', column)[1:-1])
                }
                out.loc[coords] = np.expand_dims(
                    values[column].values,
                    axis=tuple(range(1, len(coords)+1))
                )

            return out

        ds = xr.open_dataset(file_name)

        if self.py_name in ds:
            data = ds[self.py_name]
            ds.close()

            if (
                "time" in data.dims
                and list(self.coords).sort() == list(data.dims[1:]).sort()
            ):
                return data

        ds.close()
        return None



"""
The VensimFile class allows reading the original Vensim model file,
parsing it into Section elements using the FileSectionsVisitor,
parsing its sketch using SketchVisitor in order to classify the varibales
per view. The final result can be exported to an AbstractModel class in
order to build the model in another programming language.
"""
import re
from typing import Union, List
from pathlib import Path
import warnings
import parsimonious
from collections.abc import Mapping

from ..structures.abstract_model import AbstractModel

from . import vensim_utils as vu
from .vensim_section import Section
from .vensim_utils import supported_extensions


class VensimFile():
    """
    The VensimFile class allows parsing an mdl file.
    When the object is created, the model file is automatically opened;
    unnecessary tabs, whitespaces, and linebreaks are removed; and
    the sketch is split from the model equations.

    Parameters
    ----------
    mdl_path: str or pathlib.Path
        Path to the Vensim model.

    encoding: str or None (optional)
        Encoding of the source model file. If None, the encoding will be
        read from the model, if the encoding is not defined in the model
        file it will be set to 'UTF-8'. Default is None.

    """
    def __init__(self, mdl_path: Union[str, Path],
                 encoding: Union[None, str] = None):
        self.mdl_path = Path(mdl_path)
        self.root_path = self.mdl_path.parent
        self.model_text = self._read(encoding)
        self.sketch = ""
        self.view_elements = None
        self._split_sketch()

    def __str__(self):  # pragma: no cover
        return "\nVensim model file, loaded from:\n\t%s\n" % self.mdl_path

    @property
    def _verbose(self) -> str:  # pragma: no cover
        """Get model information."""
        text = self.__str__()
        for section in self.sections:
            text += section._verbose

        return text

    @property
    def verbose(self):  # pragma: no cover
        """Print model information to standard output."""
        print(self._verbose)

    def _read(self, encoding: Union[None, str]) -> str:
        """
        Read a Vensim file and assign its content to self.model_text

        Returns
        -------
        str: model file content

        """
        # check for model extension
        if self.mdl_path.suffix.lower() not in supported_extensions:
            raise ValueError(
                "The file to translate, '%s' " % self.mdl_path
                + "is not a Vensim model. It must end with .mdl extension."
            )

        if encoding is None:
            # Try detecting the encoding from the file
            encoding = vu._detect_encoding_from_file(self.mdl_path)

        with self.mdl_path.open("r", encoding=encoding,
                                errors="ignore") as in_file:
            model_text = in_file.read()

        return model_text

    def _split_sketch(self) -> None:
        """Split model from the sketch."""
        try:
            split_model = self.model_text.split("\\\\\\---///", 1)
            self.model_text = self._clean(split_model[0])
            # remove plots section, if it exists
            self.sketch = split_model[1].split("///---\\\\\\")[0]
        except LookupError:
            pass

    def _clean(self, text: str) -> str:
        """Remove unnecessary characters."""
        return re.sub(r"[\n\t\s]+", " ", re.sub(r"\\\n\t", " ", text))

    def parse(self, parse_all: bool = True) -> None:
        """
        Parse model file with parsimonious using the grammar given in
        'parsing_grammars/file_sections.peg' and the class FileSectionsVisitor
        to visit the parsed expressions.

        This breaks the model file in VensimSections, which correspond to the
        main model section and the macros.

        Parameters
        ----------
        parse_all: bool (optional)
            If True, the VensimSection objects created will be
            automatically parsed. Otherwise, these objects will only be
            added to self.sections but not parsed. Default is True.

        """
        # get model sections (__main__ + macros)
        tree = vu.Grammar.get("file_sections").parse(self.model_text)
        self.sections = FileSectionsVisitor(tree).entries

        # main section path (Python model file)
        self.sections[0].path = self.mdl_path.with_suffix(".py")

        for section in self.sections[1:]:
            # macrots paths
            section.path = self.mdl_path.parent.joinpath(
                self._clean_file_names(section.name)[0]
                ).with_suffix(".py")

        if parse_all:
            # parse all sections
            for section in self.sections:
                # parse each section
                section.parse()

    def parse_sketch(self, subview_sep: List[str]) -> None:
        """
        Parse the sketch of the model with parsimonious using the grammar
        given in 'parsing_grammars/sketch.peg' and the class SketchVisitor
        to visit the parsed expressions.

        It will modify the views_dict of the first section, including
        the dictionary of the classification of variables by views. This,
        method should be called after calling the self.parse method.

        Parameters
        ----------
        subview_sep: list
           List of the separators to use to classify the model views in
           folders and subfolders. The sepparator must be ordered in the
           same order they appear in the view name. For example,
           if a view is named "economy:demand.exports" if
           subview_sep=[":", "."] this view's variables will be included
           in the file 'exports.py' and inside the folders economy/demand.


        """
        if self.sketch:
            sketch = list(map(
                lambda x: x.strip(),
                self.sketch.split("\\\\\\---/// ")
            ))
        else:
            warnings.warn(
                "No sketch detected. The model will be built in a "
                "single file.")
            return None

        grammar = vu.Grammar.get("sketch")
        view_elements = {}
        for module in sketch:
            for sketch_line in module.split("\n"):
                # parsed line could have information about new view name
                # or of a variable inside a view
                parsed = SketchVisitor(grammar.parse(sketch_line))

                if parsed.view_name:
                    view_name = parsed.view_name
                    view_elements[view_name] = set()

                elif parsed.variable_name:
                    view_elements[view_name].add(parsed.variable_name)

        # removes views that do not include any variable in them
        non_empty_views = {
            key: value for key, value in view_elements.items() if value
        }

        # split into subviews, if subview_sep is provided
        views_dict = {}

        if len(non_empty_views) == 1:
            warnings.warn(
                "Only a single view with no subviews was detected. The model"
                " will be built in a single file.")
            return
        elif subview_sep and any(
          sep in view for sep in subview_sep for view in non_empty_views):
            escaped_separators = list(map(lambda x: re.escape(x), subview_sep))
            for full_name, values in non_empty_views.items():
                # split the full view name using the separator and make the
                # individual parts safe file or directory names
                clean_view_parts = self._clean_file_names(
                    *re.split("|".join(escaped_separators), full_name))
                # creating a nested dict for each view.subview
                # (e.g. {view_name: {subview_name: [values]}})
                nested_dict = values

                for item in reversed(clean_view_parts):
                    nested_dict = {item: nested_dict}
                # merging the new nested_dict into the views_dict, preserving
                # repeated keys
                VensimFile._merge_nested_dicts(views_dict, nested_dict)
        else:
            # view names do not have separators or separator characters
            # not provided

            if subview_sep and not any(
              sep in view for sep in subview_sep for view in non_empty_views):
                warnings.warn(
                    "The given subview separators were not matched in "
                    "any view name.")

            for view_name, elements in non_empty_views.items():
                views_dict[self._clean_file_names(view_name)[0]] = elements

        self.sections[0].split = True
        self.sections[0].views_dict = views_dict

    def get_abstract_model(self) -> AbstractModel:
        """
        Instantiate the AbstractModel object used during building. This,
        method should be called after parsing the model (self.parse) and,
        in case you want to split the variables per views, also after
        parsing the sketch (self.parse_sketch). This automatically calls the
        get_abstract_section method from the model sections.

        Returns
        -------
        AbstractModel: AbstractModel
          Abstract Model object that can be used for building the model
          in another language.

        """
        return AbstractModel(
            original_path=self.mdl_path,
            sections=tuple(section.get_abstract_section()
                           for section in self.sections))

    @staticmethod
    def _clean_file_names(*args):
        """
        Removes special characters and makes clean file names.

        Parameters
        ----------
        *args: tuple
            Any number of strings to clean.

        Returns
        -------
        clean: list
            List containing the clean strings.

        """
        return [
            re.sub(
                r"[\W]+", "",
                name.replace(" ", "_")
            ).lstrip("0123456789")
            for name in args]

    @staticmethod
    def _merge_nested_dicts(original, to_merge):
        """
        Merge dictionaries recursively, preserving common keys.

        Parameters
        ----------
        original: dict
            Dictionary onto which the merge is executed.

        to_merge: dict or set
            Dictionary to be merged to the original_dict.

        Returns
        -------
        None

        """
        if isinstance(to_merge, Mapping):
            for key, value in to_merge.items():
                if key in original:
                    if isinstance(original[key], set):
                        # convert it into a dict
                        original[key] = {"main": original[key]}
                    VensimFile._merge_nested_dicts(original[key], value)
                else:
                    original[key] = value
        else:  # it's a set
            if "main" not in original:
                original["main"] = to_merge
            else:
                original["main"].update(to_merge)
                warnings.warn("Two views with same names but different "
                              "separators where identified. They will be "
                              "joined in a single module")


class FileSectionsVisitor(parsimonious.NodeVisitor):
    """Parse file sections"""
    def __init__(self, ast):
        self.entries = [None]
        self.visit(ast)

    def visit_main(self, n, vc):
        # main will always be stored as the first entry
        if self.entries[0] is None:
            self.entries[0] = Section(
                name="__main__",
                path=Path("."),
                section_type="main",
                params=[],
                returns=[],
                content=n.text.strip(),
                split=False,
                views_dict=None
            )
        else:
            # this is needed when macro parts are in the middle of the file
            self.entries[0].content += n.text.strip()

    def visit_macro(self, n, vc):
        self.entries.append(
            Section(
                name=vc[2].strip().lower().replace(" ", "_"),
                path=Path("."),
                section_type="macro",
                params=[
                    x.strip() for x in vc[6].split(",")] if vc[6] else [],
                returns=[
                    x.strip() for x in vc[10].split(",")] if vc[10] else [],
                content=vc[13].strip(),
                split=False,
                views_dict=None
            )
        )

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text or ""


class SketchVisitor(parsimonious.NodeVisitor):
    """Sketch visitor to save the view names and the variables in each"""
    def __init__(self, ast):
        self.variable_name = None
        self.view_name = None
        self.visit(ast)

    def visit_view_name(self, n, vc):
        self.view_name = n.text.lower()

    def visit_var_definition(self, n, vc):
        if int(vc[10]) % 2 != 0:  # not a shadow variable
            self.variable_name = vc[4].replace(" ", "_").lower()

    def generic_visit(self, n, vc):
        return "".join(filter(None, vc)) or n.text.strip() or ""



"""
cmdline parser
"""
import os
import re
from ast import literal_eval
import numpy as np
import pandas as pd
from argparse import ArgumentParser, Action

from pysd import __version__
from pysd.translators.vensim.vensim_utils import supported_extensions as\
    vensim_extensions
from pysd.translators.xmile.xmile_utils import supported_extensions as\
    xmile_extensions

docs = "https://pysd.readthedocs.io/en/master/command_line_usage.html"

parser = ArgumentParser(
    description='Simulating System Dynamics Models in Python',
    prog='PySD')


#########################
# functions and actions #
#########################

def check_output(string):
    """
    Checks that out put file ends with .tab or .csv

    """
    if not string.endswith(('.tab', '.csv', '.nc')):
        parser.error(
            f'when parsing {string}'
            '\nThe output file name must be .tab, .csv or .nc...')

    return string


def check_model(string):
    """
    Checks that model file ends with .py .mdl or .xmile and that exists.

    """
    suffixes = [".py"] + vensim_extensions + xmile_extensions
    if not any(string.lower().endswith(suffix) for suffix in suffixes):
        parser.error(
            f"when parsing {string} \nThe model file name must be a Vensim"
            f" ({', '.join(vensim_extensions)}), a Xmile "
            f"({', '.join(xmile_extensions)}) or a PySD (.py) model file...")

    if not os.path.isfile(string):
        parser.error(
            f"when parsing {string}"
            "\nThe model file does not exist...")

    return string


def check_data_file(string):
    """
    Check that data file is a tab or csv file and that exists.
    """
    if not string.endswith(('.tab', '.csv', '.nc')):
        parser.error(
            f'when parsing {string}'
            '\nThe data file name must be .tab, .csv or .nc...')
    elif not os.path.isfile(string):
        parser.error(
            f'when parsing {string}'
            '\nThe data file does not exist...')
    else:
        return string


def split_files(string):
    """
    Splits the data files and returns and error if file doesn't exists
    --data 'file1.tab, file2.csv' -> ['file1.tab', 'file2.csv']
    --data file1.tab -> ['file1.tab']

    """
    return [check_data_file(s.strip()) for s in string.split(',')]


def split_columns(string):
    """
    Splits the return-columns argument or reads it from .txt
    --return-columns 'temperature c, "heat$"' -> ['temperature c', '"heat$"']
    --return-columns my_vars.txt -> ['temperature c', '"heat$"']

    """
    if string.endswith('.txt'):
        with open(string, 'r') as file:
            return [col.rstrip('\n').strip() for col in file]

    return [s.strip() for s in string.split(',')]


def split_timestamps(string):
    """
    Splits the return-timestamps argument
    --return-timestamps '1, 5, 15' -> array([1., 5., 15.])

    """
    try:
        return np.array([s.strip() for s in string.split(',')], dtype=float)
    except Exception:
        # error
        raise parser.error(
            f'when parsing {string}'
            '\nThe return time stamps must be separated by commas...\n'
            f'See {docs} for examples.')


def split_vars(string):
    """
    Splits the arguments from new_values.
    'a=5' -> {'a': ('param', 5.)}
    'b=[[1,2],[1,10]]' -> {'b': ('param', pd.Series(index=[1,2], data=[1,10]))}
    'a:5' -> {'a': ('initial', 5.)}

    """
    try:
        if '=' in string:
            # new variable value
            var, value = string.split('=')
            type = 'param'

        if ':' in string:
            # initial time value
            var, value = string.split(':')
            type = 'initial'

        if re.match(r"^[+-]?(\d*\.)?\d+$", value.strip()):
            # value is a number
            return {var.strip(): (type, float(value))}

        # value is series
        assert type == 'param'
        value = literal_eval(value)
        assert len(value) == 2
        assert len(value[0]) == len(value[1])
        return {var.strip(): (type,
                              pd.Series(index=value[0], data=value[1]))}

    except Exception:
        # error
        raise parser.error(
            f'when parsing {string}'
            '\nYou must use variable=new_value to redefine values or '
            'variable:initial_value to define initial value.'
            'variable must be a model component, new_value can be a '
            'float or a list of two list, initial_value must be a float'
            '...\n'
            f'See {docs} for examples.')


class SplitVarsAction(Action):
    """
    Convert the list of split variables from new_values to a dictionary.
    [{'a': 5.}, {'b': pd.Series(index=[1, 2], data=[1, 10])}] ->
        {'a': 5., 'b': pd.Series(index=[1, 2], data=[1, 10])}
    """
    def __init__(self, option_strings, dest, **kwargs):
        super().__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        main_dict = {'param': {}, 'initial': {}}
        for var in values:
            for var_name, (type, val) in var.items():
                main_dict[type][var_name] = val
        setattr(namespace, self.dest, main_dict)


###########
# options #
###########

parser.add_argument(
    '-v', '--version',
    action='version', version=f'PySD {__version__}')

parser.add_argument(
    '-o', '--output-file', dest='output_file',
    type=check_output, metavar='FILE',
    help='output file to save run outputs (.tab, .csv or .nc)')

parser.add_argument(
    '-p', '--progress', dest='progress',
    action='store_true', default=False,
    help='show progress bar during model integration')

parser.add_argument(
    '-r', '--return-columns', dest='return_columns',
    action='store', type=split_columns,
    metavar='\'var1, var2, .., varN\' or FILE (.txt)',
    help='provide the return columns separated by commas or a .txt file'
         ' where each row is a variable')

parser.add_argument(
    '-e', '--export', dest='export_file',
    type=str, metavar='FILE',
    help='export to a pickle stateful objects states at the end of the '
         'simulation')

parser.add_argument(
    '-i', '--import-initial', dest='import_file',
    type=str, metavar='FILE',
    help='import stateful objects states from a pickle file,'
         'if given initial conditions from var:value will be ignored')


###################
# Model arguments #
###################

model_arguments = parser.add_argument_group(
    'model arguments',
    'Modify model control variables.')

model_arguments.add_argument(
    '-I', '--initial-time', dest='initial_time',
    action='store', type=float, metavar='VALUE',
    help='modify initial time of the simulation')

model_arguments.add_argument(
    '-F', '--final-time', dest='final_time',
    action='store', type=float, metavar='VALUE',
    help='modify final time of the simulation')

model_arguments.add_argument(
    '-T', '--time-step', dest='time_step',
    action='store', type=float, metavar='VALUE',
    help='modify time step of the simulation')

model_arguments.add_argument(
    '-S', '--saveper', dest='saveper',
    action='store', type=float, metavar='VALUE',
    help='modify time step of the output')

model_arguments.add_argument(
    '-R', '--return-timestamps', dest='return_timestamps',
    action='store', type=split_timestamps,
    metavar='\'value1, value2, .., valueN\'',
    help='provide the return time stamps separated by commas, if given '
         '--saveper will be ignored')

model_arguments.add_argument(
    '-D', '--data', dest='data_files',
    action='store', type=split_files, metavar='\'FILE1, FILE2, .., FILEN\'',
    help='input data file or files to run the model')

#########################
# Translation arguments #
#########################

trans_arguments = parser.add_argument_group(
    'translation arguments',
    'Configure the translation of the original model.')

trans_arguments.add_argument(
    '--translate', dest='run',
    action='store_false', default=True,
    help='only translate the model_file, '
         'it does not run it after translation')

trans_arguments.add_argument(
    '--split-views', dest='split_views',
    action='store_true', default=False,
    help='parse the sketch to detect model elements in each model view,'
         ' and then translate each view in a separate Python file')

trans_arguments.add_argument(
    '--subview-sep', dest='subview_sep',
    action='store', nargs="*", default=[],
    metavar='separator_1 separator_2 ... separator_n',
    help='further division of views into subviews, by identifying the '
         'separator string in the view name, only availabe if --split-views'
         ' is used. Passing positional arguments after this argument will'
         ' not work')


#######################
# Warnings and errors #
#######################

warn_err_arguments = parser.add_argument_group(
    'warning and errors arguments',
    'Modify warning and errors management.')

warn_err_arguments.add_argument(
    '--missing-values', dest='missing_values', default="warning",
    action='store', type=str, choices=['warning', 'raise', 'ignore', 'keep'],
    help='exception with missing values, \'warning\' (default) shows a '
         'warning message and interpolates the values, \'raise\' raises '
         'an error, \'ignore\' interpolates the values without showing '
         'anything, \'keep\' keeps the missing values')


########################
# Positional arguments #
########################

parser.add_argument('model_file', metavar='model_file', type=check_model,
                    help='Vensim, Xmile or PySD model file')

parser.add_argument('new_values',
                    metavar='variable=new_value', type=split_vars,
                    nargs='*', action=SplitVarsAction,
                    help='redefine the value of variable with new value. '
                    'variable must be a model component, new_value may be a '
                    'float or a list of two lists')

# The destination new_values2 will never be used as the previous argument
# is given also with nargs='*'. Nevertheless, the following variable
# is declared for documentation
parser.add_argument('new_values2',
                    metavar='variable:initial_value', type=split_vars,
                    nargs='*', action=SplitVarsAction,
                    help='redefine the initial value of variable. '
                    'variable must be a model stateful element, initial_value'
                    ' must be a float')


#########
# Usage #
#########

parser.usage = parser.format_usage().replace("usage: PySD", "python -m pysd")



"""
The ModelBuilder class allows converting the AbstractModel into a
PySD model writing the Python code in files that can be loaded later
with PySD Model class. Each Abstract level has its own Builder. However,
the user is only required to create a ModelBuilder object using the
AbstractModel and call the `build_model` method.
"""
from warnings import warn
import textwrap
import black
import json
from pathlib import Path
from typing import Union

from pysd.translators.structures.abstract_model import\
    AbstractComponent, AbstractElement, AbstractControlElement,\
    AbstractModel, AbstractSection

from . import python_expressions_builder as vs
from .namespace import NamespaceManager
from .subscripts import SubscriptManager
from .imports import ImportsManager
from pysd._version import __version__


class ModelBuilder:
    """
    ModelBuilder allows building a PySD Python model from the
    Abstract Model.

    Parameters
    ----------
    abstract_model: AbstractModel
        The abstract model to build.

    """

    def __init__(self, abstract_model: AbstractModel):
        self.__dict__ = abstract_model.__dict__.copy()
        # load sections
        self.sections = [
            SectionBuilder(section)
            for section in abstract_model.sections
        ]
        # create the macrospace (namespace of macros)
        self.macrospace = {
            section.name: section for section in self.sections[1:]}

    def build_model(self) -> Path:
        """
        Build the Python model in a file callled as the orginal model
        but with '.py' suffix.

        Returns
        -------
        path: pathlib.Path
            The path to the new PySD model.

        """
        for section in self.sections:
            # add macrospace information to each section and build it
            section.macrospace = self.macrospace
            section.build_section()

        # return the path to the main file
        return self.sections[0].path


class SectionBuilder:
    """
    SectionBuilder allows building a section of the PySD model. Each
    section will be a file unless the model has been set to be split
    in modules.

    Parameters
    ----------
    abstract_section: AbstractSection
        The abstract section to build.

    """
    def __init__(self, abstract_section: AbstractSection):
        self.__dict__ = abstract_section.__dict__.copy()
        self.root = self.path.parent  # the folder where the model is
        self.model_name = self.path.with_suffix("").name  # name of the model
        # Create subscript manager object with subscripts_dict
        self.subscripts = SubscriptManager(
            abstract_section.subscripts, self.root)
        # Load the elements in the section
        self.elements = [
            ElementBuilder(element, self)
            for element in abstract_section.elements
        ]
        # Create the namespace of the section
        self.namespace = NamespaceManager(self.params)
        # Create an imports manager
        self.imports = ImportsManager()
        # Create macrospace (namespace of macros)
        self.macrospace = {}
        # Create parameters dict necessary in macros
        self.params = {
            key: self.namespace.namespace[key]
            for key in self.params
        }
        # Import xarray if there are any subscripts defined in the section
        if self.subscripts.subscripts:
            self.imports.add("xarray")

    def build_section(self) -> None:
        """
        Build the Python section in a file callled as the orginal model
        if the section is main or in a file called as the macro name
        if the section is a macro.
        """
        # Firts iteration over elements to recover their information
        for element in self.elements:
            # Add element to namespace
            self.namespace.add_to_namespace(element.name)
            identifier = self.namespace.namespace[element.name]
            element.identifier = identifier
            # Add element subscripts information to the subscript manager
            self.subscripts.elements[identifier] = element.subscripts

        # Build elements
        for element in self.elements:
            element.build_element()

        if self.split:
            # Build modular section
            self._build_modular(self.views_dict)
        else:
            # Build one-file section
            self._build()

    def _process_views_tree(self, view_name: str,
                            view_content: Union[dict, set],
                            wdir: Path) -> dict:
        """
        Creates a directory tree based on the elements_per_view dictionary.
        If it's the final view, it creates a file, if not, it creates a folder.
        """
        if isinstance(view_content, set):
            # Will become a module
            # Convert subview elements names to Python names
            view_content = {
                self.namespace.cleanspace[var] for var in view_content
            }

            # Get subview elements (ordered)
            subview_elems = sorted(
                [
                    element for element in self.elements_remaining
                    if element.identifier in view_content
                    and not element.control_var
                ],
                key=lambda x: x.identifier)

            # Get the names of the elements and include their
            # information in the elements_added dictionary
            subview_elems_names = [
                element.identifier for element in subview_elems
            ]
            view_path = ".".join(view_name.parts[1:])
            self.elements_added.update({
                var: view_path for var in subview_elems_names
            })

            if len(view_content) != len(subview_elems_names):
                # Some elements from the view where not added
                for var in view_content.difference(subview_elems_names):
                    original_name = self.namespace.get_original_name(var)
                    if var in self.elements_added:
                        # Element already added in another view
                        warn(
                            f"Variable '{original_name}' is declared as "
                            f"a workbench variable in '{view_path}' but "
                            "it has been already added in "
                            f"'{self.elements_added[var]}'."
                        )
                    else:
                        # Element is a control variable
                        warn(
                            f"Control variable '{original_name}' is "
                            "declared as a workbench variable in "
                            f"'{view_path}'. As it is a control "
                            "variable, this declaration will be ignored "
                            "and added to the main module only."
                        )

            # Remove elements from remaining ones
            [
                self.elements_remaining.remove(element)
                for element in subview_elems
            ]

            if subview_elems:
                # Build the module (only when they are variables)
                self._build_separate_module(subview_elems, view_name, wdir)

            return list(subview_elems_names)
        else:
            # The current view has subviews
            (wdir / view_name).mkdir(exist_ok=True)
            subviews = {
                subview_name: self._process_views_tree(
                    view_name / subview_name, subview_content, wdir)
                for subview_name, subview_content in view_content.items()
            }
            # Avoid includying empty views to the dictionary
            return {
                subview_name: subview_content
                for subview_name, subview_content in subviews.items()
                if subview_content
            }

    def _build_modular(self, elements_per_view: dict) -> None:
        """ Build modular section """
        self.elements_remaining = self.elements.copy()
        self.elements_added = {}
        elements_per_view = self._process_views_tree(
            Path("modules_" + self.model_name), elements_per_view, self.root)

        for element in self.elements_remaining:
            if not element.control_var:
                warn(
                    f"Variable '{element.name}' is not declared as a "
                    "workbench variable in any view. It will be added to "
                    "the main module."
                )
        # Building main file using the build function
        self._build_main_module(self.elements_remaining)

        # Build subscripts dir and moduler .json files
        for file, values in {
          "modules_%s/_modules": elements_per_view,
          "_subscripts_%s": self.subscripts.subscripts}.items():

            with self.root.joinpath(
                file % self.model_name).with_suffix(
                    ".json").open("w") as outfile:
                json.dump(values, outfile, indent=4, sort_keys=True)

    def _build_separate_module(self, elements: list, module_name: str,
                               module_dir: str) -> None:
        """
        Constructs and writes the Python representation of a specific model
        module, when the split_views=True in the read_vensim function.

        Parameters
        ----------
        elements: list
            Elements belonging to the module module_name.

        module_name: str
            Name of the module

        module_dir: str
            Path of the directory where module files will be stored.

        Returns
        -------
        None

        """
        text = textwrap.dedent('''
        """
        Module %(module_name)s
        Translated using PySD version %(version)s
        """
        ''' % {
            "module_name": ".".join(module_name.parts[1:]),
            "version": __version__,
        })
        funcs = self._generate_functions(elements)
        text += funcs
        text = black.format_file_contents(
            text, fast=True, mode=black.FileMode())

        outfile_name = module_dir / module_name.with_suffix(".py")

        with outfile_name.open("w", encoding="UTF-8") as out:
            out.write(text)

    def _build_main_module(self, elements: list) -> None:
        """
        Constructs and writes the Python representation of the main model
        module, when the split_views=True in the read_vensim function.

        Parameters
        ----------
        elements: list
            Elements belonging to the main module. Ideally, there should
            only be the initial_time, final_time, saveper and time_step,
            functions, though there might be others in some situations.
            Each element is a dictionary, with the various components
            needed to assemble a model component in Python syntax. This
            will contain multiple entries for elements that have multiple
            definitions in the original file, and which need to be combined.

        Returns
        -------
        None

        """
        # separating between control variables and rest of variables
        control_vars, funcs = self._build_variables(elements)

        self.imports.add("utils", "load_model_data")
        self.imports.add("utils", "load_modules")

        # import of needed functions and packages
        text = self.imports.get_header(self.path.name)

        # import subscript dict from json file
        text += textwrap.dedent("""
        __pysd_version__ = '%(version)s'

        __data = {
            'scope': None,
            'time': lambda: 0
        }

        _root = Path(__file__).parent
        %(params)s
        _subscript_dict, _modules = load_model_data(
            _root, "%(model_name)s")

        component = Component()
        """ % {
            "params": f"\n        _params = {self.params}\n"
                      if self.params else "",
            "model_name": self.model_name,
            "version": __version__
        })

        text += self._get_control_vars(control_vars)

        text += textwrap.dedent("""
            # load modules from modules_%(model_name)s directory
            exec(load_modules("modules_%(model_name)s", _modules, _root, []))

            """ % {
                "model_name": self.model_name,
            })

        text += funcs
        text = black.format_file_contents(
            text, fast=True, mode=black.FileMode())

        with self.path.open("w", encoding="UTF-8") as out:
            out.write(text)

    def _build(self) -> None:
        """
        Constructs and writes the Python representation of a section.

        Returns
        -------
        None

        """
        control_vars, funcs = self._build_variables(self.elements)

        text = self.imports.get_header(self.path.name)
        indent = "\n        "
        # Generate params dict for macro parameters
        params = f"{indent}_params = {self.params}\n"\
            if self.params else ""
        # Generate subscripts dir
        subs = f"{indent}_subscript_dict = {self.subscripts.subscripts}"\
            if self.subscripts.subscripts else ""

        text += textwrap.dedent("""
        __pysd_version__ = '%(version)s'

        __data = {
            'scope': None,
            'time': lambda: 0
        }

        _root = Path(__file__).parent
        %(params)s
        %(subscript_dict)s

        component = Component()
        """ % {
            "subscript_dict": subs,
            "params": params,
            "version": __version__,
        })

        text += self._get_control_vars(control_vars) + funcs

        text = black.format_file_contents(
            text, fast=True, mode=black.FileMode())

        with self.path.open("w", encoding="UTF-8") as out:
            out.write(text)

    def _build_variables(self, elements: dict) -> tuple:
        """
        Build model variables (functions) and separate then in control
        variables and regular variables.

        Returns
        -------
        control_vars, regular_vars: tuple, str
            control_vars is a tuple of length 2. First element is the
            dictionary of original control vars. Second is the string to
            add the control variables' functions. regular_vars is the
            string to add the regular variables' functions.

        """
        # returns of the control variables
        control_vars_dict = {
            "initial_time": "__data['time'].initial_time()",
            "final_time": "__data['time'].final_time()",
            "time_step": "__data['time'].time_step()",
            "saveper": "__data['time'].saveper()"
        }
        regular_vars = []
        control_vars = []

        for element in elements:
            if element.identifier in control_vars_dict:
                # change the return expression in the element and update
                # the dict with the original expression
                control_vars_dict[element.identifier], element.expression =\
                    element.expression, control_vars_dict[element.identifier]
                control_vars.append(element)
            else:
                regular_vars.append(element)

        if len(control_vars) == 0:
            # macro objects, no control variables
            control_vars_dict = ""
        else:
            control_vars_dict = """
        _control_vars = {
            "initial_time": lambda: %(initial_time)s,
            "final_time": lambda: %(final_time)s,
            "time_step": lambda: %(time_step)s,
            "saveper": lambda: %(saveper)s
        }
        """ % control_vars_dict

        return (control_vars_dict,
                self._generate_functions(control_vars)),\
            self._generate_functions(regular_vars)

    def _generate_functions(self, elements: dict) -> str:
        """
        Builds all model elements as functions in string format.
        NOTE: this function calls the build_element function, which
        updates the import_modules.
        Therefore, it needs to be executed before the method
        _generate_automatic_imports.

        Parameters
        ----------
        elements: dict
            Each element is a dictionary, with the various components
            needed to assemble a model component in Python syntax. This
            will contain multiple entries for elements that have multiple
            definitions in the original file, and which need to be combined.

        Returns
        -------
        funcs: str
            String containing all formated model functions

        """
        return "\n".join(
            [element._build_element_out() for element in elements]
        )

    def _get_control_vars(self, control_vars: str) -> str:
        """
        Create the section of control variables

        Parameters
        ----------
        control_vars: str
            Functions to define control variables.

        Returns
        -------
        text: str
            Control variables section and header of model variables section.

        """
        text = textwrap.dedent("""
        #######################################################################
        #                          CONTROL VARIABLES                          #
        #######################################################################
        %(control_vars_dict)s
        def _init_outer_references(data):
            for key in data:
                __data[key] = data[key]


        @component.add(name="Time")
        def time():
            '''
            Current time of the model.
            '''
            return __data['time']()

        """ % {"control_vars_dict": control_vars[0]})

        text += control_vars[1]

        text += textwrap.dedent("""
        #######################################################################
        #                           MODEL VARIABLES                           #
        #######################################################################
        """)

        return text


class ElementBuilder:
    """
    ElementBuilder allows building an element of the PySD model.

    Parameters
    ----------
    abstract_element: AbstractElement
        The abstract element to build.
    section: SectionBuilder
        The section where the element is defined. Necessary to give the
        acces to the subscripts and namespace.

    """
    def __init__(self, abstract_element: AbstractElement,
                 section: SectionBuilder):
        self.__dict__ = abstract_element.__dict__.copy()
        self.control_var = isinstance(abstract_element, AbstractControlElement)
        # Set element type and subtype to None
        self.type = None
        self.subtype = None
        # Get the arguments of the element
        self.arguments = getattr(self.components[0], "arguments", "")
        # Load the components of the element
        self.components = [
            ComponentBuilder(component, self, section)
            for component in abstract_element.components
        ]
        self.section = section
        # Get the subscripts of the element after merging all the components
        self.subscripts = section.subscripts.make_merge_list(
            [component.subscripts[0] for component in self.components])
        # Get the subscript dictionary of the element
        self.subs_dict = section.subscripts.make_coord_dict(self.subscripts)
        # Dictionaries to save dependencies and objects related to the element
        self.dependencies = {}
        self.other_dependencies = {}
        self.objects = {}

    def build_element(self) -> None:
        """
        Build the element. Returns the string to include in the section which
        will be a decorated function definition and possible objects.
        """
        # TODO think better how to build the components at once to build
        # in one declaration the external objects
        # TODO include some kind of magic vectorization to identify patterns
        # that can be easily vecorized (GET, expressions, Stocks...)

        # Build the components of the element
        [component.build_component() for component in self.components]
        expressions = []
        for component in self.components:
            expr, subs, except_subscripts = component.get()
            if expr is None:
                # The expr is None when the component has been "added"
                # to an existing object using the add method
                continue
            if isinstance(subs, list):
                # Get the list of locs for the component
                # Subscripts dict will be a list when the component is
                # translated to an object that groups may components
                # via 'add' method.
                loc = [vs.visit_loc(subsi, self.subs_dict, True)
                       for subsi in subs]
            else:
                # Get the loc of the component
                loc = vs.visit_loc(subs, self.subs_dict, True)

            # Get the locs of the :EXCLUDE: parameters if any
            exc_loc = [
                vs.visit_loc(subs_e, self.subs_dict, True)
                for subs_e in except_subscripts
            ]
            expressions.append({
                "expr": expr,
                "subs": subs,
                "loc": loc,
                "loc_except": exc_loc
            })

        if len(expressions) > 1:
            # NUMPY: xrmerge would be sustitute by a multiple line definition
            # e.g.:
            # value = np.empty((len(dim1), len(dim2)))
            # value[:, 0] = expression1
            # value[:, 1] = expression2
            # return value
            # This allows reference to the same variable
            # from: VAR[A] = 5; VAR[B] = 2*VAR[A]
            # to: value[0] = 5; value[1] = 2*value[0]
            self.section.imports.add("numpy")
            self.pre_expression =\
                "value = xr.DataArray(np.nan, {%s}, %s)\n" % (
                    ", ".join("'%(dim)s': _subscript_dict['%(dim)s']" %
                              {"dim": subs} for subs in self.subscripts),
                    self.subscripts)
            for expression in expressions:
                # Generate the pre_expression, operations to compute in
                # the body of the function
                if expression["expr"].subscripts:
                    # Get the values
                    # NUMPY not necessary
                    expression["expr"].lower_order(-1)
                    expression["expr"].expression += ".values"
                if expression["loc_except"]:
                    # There is an excep in the definition of the component
                    self.pre_expression += self._manage_except(expression)
                elif isinstance(expression["subs"], list):
                    # There are mixed definitions which include multicomponent
                    # object
                    self.pre_expression += self._manage_multi_def(expression)
                else:
                    # Regular loc for a component
                    self.pre_expression +=\
                        "value.loc[%(loc)s] = %(expr)s\n" % expression

            # Return value
            self.expression = "value"
        else:
            self.pre_expression = ""
            # NUMPY: reshape to the final shape if needed)
            # expressions[0]["expr"].reshape(self.section.subscripts, {})
            if not expressions[0]["expr"].subscripts and self.subscripts:
                # Updimension the return value to an array
                self.expression = "xr.DataArray(%s, %s, %s)\n" % (
                     expressions[0]["expr"],
                     self.section.subscripts.simplify_subscript_input(
                         self.subs_dict)[1],
                     list(self.subs_dict)
                )
            else:
                # Return the expression
                self.expression = expressions[0]["expr"]

        # Merge the types of the components (well defined element should
        # have only one type and subtype)
        self.type = ", ".join(
            set(component.type for component in self.components)
        )
        self.subtype = ", ".join(
            set(component.subtype for component in self.components)
        )
        if ", " in self.type:
            warn(
                f"Variable '{self.name}' is defined with different types:"
                f" '{self.type}'. This may cause bugs when trying to "
                "change its value or applying other methods from the "
                "pysd.py_backend.model.Model class. Running the model "
                "without modifying this variable should not cause any "
                "bug."
            )
        elif ", " in self.subtype:
            warn(
                f"Variable '{self.name}' is defined with different subtypes:"
                f" '{self.subtype}'. This may cause bugs when trying to "
                "change its value or applying other methods from the "
                "pysd.py_backend.model.Model class. Running the model "
                "without modifying this variable should not cause any "
                "bug."
            )

    def _manage_multi_def(self, expression: dict) -> str:
        """
        Manage multiline definitions when some of them (not all) are
        merged to one object.
        """
        final_expr = "def_subs = xr.zeros_like(value, dtype=bool)\n"
        for loc in expression["loc"]:
            # coordinates of the object
            final_expr += f"def_subs.loc[{loc}] = True\n"

        # replace the values matching the coordinates
        return final_expr + "value.values[def_subs.values] = "\
            "%(expr)s[def_subs.values]\n" % expression

    def _manage_except(self, expression: dict) -> str:
        """
        Manage except declarations by not asigning its values.
        """
        if expression["subs"] == self.subs_dict:
            # Final subscripts are the same as the main subscripts
            # of the component. Generate a True array like value
            final_expr = "except_subs = xr.ones_like(value, dtype=bool)\n"
        else:
            # Final subscripts are greater than the main subscripts
            # of the component. Generate a False array like value and
            # set to True the subarray of the component coordinates
            final_expr = "except_subs = xr.zeros_like(value, dtype=bool)\n"\
                         "except_subs.loc[%(loc)s] = True\n" % expression

        for except_subs in expression["loc_except"]:
            # We set to False the dimensions in the EXCEPT
            final_expr += "except_subs.loc[%s] = False\n" % except_subs

        if expression["expr"].subscripts:
            # Assign the values of an array
            if expression["subs"] == self.subs_dict:
                return final_expr + "value.values[except_subs.values] = "\
                    "%(expr)s[except_subs.values]\n" % expression
            else:
                return final_expr + "value.values[except_subs.values] = "\
                    "%(expr)s[except_subs.loc[%(loc)s].values]\n" % expression
        else:
            # Assign the values of a float
            return final_expr + "value.values[except_subs.values] = "\
                "%(expr)s\n" % expression

    def _build_element_out(self) -> str:
        """
        Returns a string that has processed a single element dictionary.

        Returns
        -------
        func: str
            The function to write in the model file.

        """
        # Contents of the function (body + return)
        contents = self.pre_expression + "return %s" % self.expression

        # Get the objects to create as string
        objects = "\n\n".join([
            value["expression"] for value in self.objects.values()
            if value["expression"] is not None
        ])

        # Format the limits to get them as a string
        self.limits = self._format_limits(self.limits)

        # Update arguments with final subs to alllow passing arguments
        # with subscripts to the lookups
        if self.arguments == 'x':
            self.arguments = 'x, final_subs=None'

        # Define variable metadata for the @component decorator
        self.name = repr(self.name)
        meta_data = ["name=%(name)s"]

        # Include basic metadata (units, limits, dimensions)
        if self.units:
            meta_data.append("units=%(units)s")
            self.units = repr(self.units)
        if self.limits:
            meta_data.append("limits=%(limits)s")
        if self.subscripts:
            self.section.imports.add("subs")
            meta_data.append("subscripts=%(subscripts)s")

        # Include component type and subtype
        meta_data.append("comp_type='%(type)s'")
        meta_data.append("comp_subtype='%(subtype)s'")

        # Include dependencies
        if self.dependencies:
            meta_data.append("depends_on=%(dependencies)s")
        if self.other_dependencies:
            meta_data.append("other_deps=%(other_dependencies)s")

        # Get metadata decorator
        self.meta_data = f"@component.add({', '.join(meta_data)})"\
            % self.__dict__

        # Clean the documentation and add it to the beggining of contents
        if self.documentation:
            doc = self.documentation.replace("\\", "\n")
            contents = f'"""\n{doc}\n"""\n'\
                + contents

        indent = 12

        # Convert newline indicator and add expected level of indentation
        self.contents = contents.replace("\n", "\n" + " " * (indent+4))
        self.objects = objects.replace("\n", "\n" + " " * indent)

        # Return the decorated function definition with the object declarations
        return textwrap.dedent('''
            %(meta_data)s
            def %(identifier)s(%(arguments)s):
                %(contents)s


            %(objects)s
            ''' % self.__dict__)

    def _format_limits(self, limits: tuple) -> str:
        """Format the limits of an element to print them properly"""
        if limits == (None, None):
            return None

        new_limits = []
        for value in limits:
            value = repr(value)
            if value == "nan" or value == "None":
                # add numpy.nan to the values
                self.section.imports.add("numpy")
                new_limits.append("np.nan")
            elif value.endswith("inf"):
                # add numpy.inf to the values
                self.section.imports.add("numpy")
                new_limits.append(value.strip("inf") + "np.inf")
            else:
                # add numeric value
                new_limits.append(value)

        if new_limits[0] == "np.nan" and new_limits[1] == "np.nan":
            # if both are numpy.nan do not include limits
            return None

        return "(" + ", ".join(new_limits) + ")"


class ComponentBuilder:
    """
    ComponentBuilder allows building a component of the PySD model.

    Parameters
    ----------
    abstract_component: AbstracComponent
        The abstract component to build.
    element: ElementBuilder
        The element where the component is defined. Necessary to give the
        acces to the merging subscripts and other components.
    section: SectionBuilder
        The section where the element is defined. Necessary to give the
        acces to the subscripts and namespace.

    """
    def __init__(self, abstract_component: AbstractComponent,
                 element: ElementBuilder, section: SectionBuilder):
        self.__dict__ = abstract_component.__dict__.copy()
        self.element = element
        self.section = section
        if not hasattr(self, "keyword"):
            self.keyword = None

    def build_component(self) -> None:
        """
        Build model component parsing the Abstract Syntax Tree.
        """
        self.subscripts_dict = self.section.subscripts.make_coord_dict(
            self.subscripts[0])
        self.except_subscripts = [self.section.subscripts.make_coord_dict(
            except_list) for except_list in self.subscripts[1]]
        self.ast_build = vs.ASTVisitor(self).visit()

    def get(self) -> tuple:
        """
        Get build component to build the element.

        Returns
        -------
        ast_build: BuildAST
            Parsed AbstractSyntaxTree.
        subscript_dict: dict or list of dicts
            The subscripts of the component.
        except_subscripts: list of dicts
            The subscripts to avoid.

        """
        return self.ast_build, self.subscripts_dict, self.except_subscripts



"""
The following abstract structures are used to build the Abstract Syntax
Tree (AST). In general, there is no hierarchy between them. For example,
an ArithmeticStructure can contain a CallStructure which at the same time
contains another ArithmeticStructure. However, some of them could not be
inside another structures due to the restrictions of the source languages.
For example, the GetConstantsStructure cannot be a part of another structure
because it has to appear after the '=' sign in Vensim and not be followed by
anything else.
"""
from dataclasses import dataclass
from typing import Union, Optional


class AbstractSyntax:
    """
    Generic class. All Abstract Synax structured are childs of that class.
    Used for typing.
    """
    pass


@dataclass
class ArithmeticStructure(AbstractSyntax):
    """
    Dataclass for an arithmetic structure.

    Parameters
    ----------
    operators: list
        List of operators applied between the arguments
    arguments: list
        The arguments of the arithmetics operations.

    """
    operators: list
    arguments: list

    def __str__(self) -> str:  # pragma: no cover
        return "ArithmeticStructure:\n\t %s %s" % (
            self.operators, self.arguments)


@dataclass
class LogicStructure(AbstractSyntax):
    """
    Dataclass for a logic structure.

    Parameters
    ----------
    operators: list
        List of operators applied between the arguments
    arguments: list
        The arguments of the logic operations.

    """
    operators: list
    arguments: list

    def __str__(self) -> str:  # pragma: no cover
        return "LogicStructure:\n\t %s %s" % (
            self.operators, self.arguments)


@dataclass
class SubscriptsReferenceStructure(AbstractSyntax):
    """
    Dataclass for a subscript reference structure.

    Parameters
    ----------
    subscripts: tuple
        The list of subscripts referenced.

    """
    subscripts: tuple

    def __str__(self) -> str:  # pragma: no cover
        return "SubscriptReferenceStructure:\n\t %s" % self.subscripts


@dataclass
class ReferenceStructure(AbstractSyntax):
    """
    Dataclass for an element reference structure.

    Parameters
    ----------
    reference: str
        The name of the referenced element.
    subscripts: SubscriptsReferenceStructure or None
        The subscrips used in the reference.

    """
    reference: str
    subscripts: Union[SubscriptsReferenceStructure, None] = None

    def __str__(self) -> str:  # pragma: no cover
        return "ReferenceStructure:\n\t %s%s" % (
            self.reference,
            "\n\t" + str(self.subscripts or "").replace("\n", "\n\t"))


@dataclass
class CallStructure(AbstractSyntax):
    """
    Dataclass for a call structure.

    Parameters
    ----------
    function: ReferenceStructure
        The reference of the callable.
    arguments: tuple
        The list of arguments used for calling the function.

    """
    function: ReferenceStructure
    arguments: tuple

    def __str__(self) -> str:  # pragma: no cover
        return "CallStructure:\n\t%s(%s)" % (
            self.function,
            "\n\t\t,".join([
                "\n\t\t" + str(arg).replace("\n", "\n\t\t")
                for arg in self.arguments
            ]))


@dataclass
class GameStructure(AbstractSyntax):
    """
    Dataclass for a game structure.

    Parameters
    ----------
    expression: AST
        The expression inside the game call.

    """
    expression: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "GameStructure:\n\t%s" % self.expression


@dataclass
class AllocateAvailableStructure(AbstractSyntax):
    """
    Dataclass for a Allocate Available structure.

    Parameters
    ----------
    request: AbstractSyntax
        The reference to the request variable.
    pp: AbstractSyntax
        The reference to the priority variable.
    avail: AbstractSyntax or float
        The total available supply.

    """
    request: AbstractSyntax
    pp: AbstractSyntax
    avail: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "AllocateAvailableStructure:\n\t%s,\n\t%s,\n\t%s" % (
            self.request, self.pp, self.avail
        )


@dataclass
class AllocateByPriorityStructure(AbstractSyntax):
    """
    Dataclass for a Allocate By Priority structure.

    Parameters
    ----------
    request: AbstractSyntax
        The reference to the request variable.
    priority: AbstractSyntax
        The reference to the priority variable.
    size: AbstractSyntax or int
        The size of the last dimension.
    width: AbstractSyntax or float
        The width between priorities.
    supply: AbstractSyntax or float
        The total supply.

    """
    request: AbstractSyntax
    priority: AbstractSyntax
    size: Union[AbstractSyntax, int]
    width: Union[AbstractSyntax, float]
    supply: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "AllocateByPriorityStructure:"\
               "\n\t%s,\n\t%s,\n\t%s,\n\t%s,\n\t%s" % (
                   self.request, self.priority, self.size,
                   self.width, self.supply
                )


@dataclass
class InitialStructure(AbstractSyntax):
    """
    Dataclass for a initial structure.

    Parameters
    ----------
    initial: AST
        The expression inside the initial call.

    """
    initial: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "InitialStructure:\n\t%s" % (
            self.initial)


@dataclass
class IntegStructure(AbstractSyntax):
    """
    Dataclass for an integ/stock structure.

    Parameters
    ----------
    flow: AST
        The flow of the stock.
    initial: AST
        The initial value of the stock.
    non_negative: bool (optional)
        If True the stock cannot be negative. Default is False.

    """
    flow: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]
    non_negative: Optional[bool] = False

    def __str__(self) -> str:  # pragma: no cover
        return "IntegStructure:\n\t%s,\n\t%s" % (
            self.flow,
            self.initial)


@dataclass
class DelayStructure(AbstractSyntax):
    """
    Dataclass for a delay structure.

    Parameters
    ----------
    input: AST
        The input of the delay.
    delay_time: AST
        The delay time value of the delay.
    initial: AST
        The initial value of the delay.
    order: float
        The order of the delay.

    """
    input: Union[AbstractSyntax, float]
    delay_time: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]
    order: float

    def __str__(self) -> str:  # pragma: no cover
        return "DelayStructure (order %s):\n\t%s,\n\t%s,\n\t%s" % (
            self.order,
            self.input,
            self.delay_time,
            self.initial)


@dataclass
class DelayNStructure(AbstractSyntax):
    """
    Dataclass for a delay n structure.

    Parameters
    ----------
    input: AST
        The input of the delay.
    delay_time: AST
        The delay time value of the delay.
    initial: AST
        The initial value of the delay.
    order: float
        The order of the delay.

    """
    input: Union[AbstractSyntax, float]
    delay_time: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]
    order: Union[AbstractSyntax, float]

    # DELAY N may behave different than other delays when the delay time
    # changes during integration

    def __str__(self) -> str:  # pragma: no cover
        return "DelayNStructure (order %s):\n\t%s,\n\t%s,\n\t%s" % (
            self.order,
            self.input,
            self.delay_time,
            self.initial)


@dataclass
class DelayFixedStructure(AbstractSyntax):
    """
    Dataclass for a delay fixed structure.

    Parameters
    ----------
    input: AST
        The input of the delay.
    delay_time: AST
        The delay time value of the delay.
    initial: AST
        The initial value of the delay.

    """
    input: Union[AbstractSyntax, float]
    delay_time: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "DelayFixedStructure:\n\t%s,\n\t%s,\n\t%s" % (
            self.input,
            self.delay_time,
            self.initial)


@dataclass
class SmoothStructure(AbstractSyntax):
    """
    Dataclass for a smooth structure.

    Parameters
    ----------
    input: AST
        The input of the smooth.
    delay_time: AST
        The smooth time value of the smooth.
    initial: AST
        The initial value of the smooth.
    order: float
        The order of the smooth.

    """
    input: Union[AbstractSyntax, float]
    smooth_time: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]
    order: float

    def __str__(self) -> str:  # pragma: no cover
        return "SmoothStructure (order %s):\n\t%s,\n\t%s,\n\t%s" % (
            self.order,
            self.input,
            self.smooth_time,
            self.initial)


@dataclass
class SmoothNStructure(AbstractSyntax):
    """
    Dataclass for a smooth n structure.

    Parameters
    ----------
    input: AST
        The input of the smooth.
    delay_time: AST
        The smooth time value of the smooth.
    initial: AST
        The initial value of the smooth.
    order: float
        The order of the smooth.

    """
    input: Union[AbstractSyntax, float]
    smooth_time: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]
    order: Union[AbstractSyntax, float]

    # SMOOTH N may behave different than other smooths with RungeKutta
    # integration

    def __str__(self) -> str:  # pragma: no cover
        return "SmoothNStructure (order %s):\n\t%s,\n\t%s,\n\t%s" % (
            self.order,
            self.input,
            self.smooth_time,
            self.initial)


@dataclass
class TrendStructure(AbstractSyntax):
    """
    Dataclass for a trend structure.

    Parameters
    ----------
    input: AST
        The input of the trend.
    average_time: AST
        The average time value of the trend.
    initial_trend: AST
        The initial trend value of the trend.

    """
    input: Union[AbstractSyntax, float]
    average_time: Union[AbstractSyntax, float]
    initial_trend: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "TrendStructure:\n\t%s,\n\t%s,\n\t%s" % (
            self.input,
            self.average_time,
            self.initial)


@dataclass
class ForecastStructure(AbstractSyntax):
    """
    Dataclass for a forecast structure.

    Parameters
    ----------
    input: AST
        The input of the forecast.
    averae_time: AST
        The average time value of the forecast.
    horizon: float
        The horizon value of the forecast.
    initial_trend: AST
        The initial trend value of the forecast.

    """
    input: Union[AbstractSyntax, float]
    average_time: Union[AbstractSyntax, float]
    horizon: Union[AbstractSyntax, float]
    initial_trend: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "ForecastStructure:\n\t%s,\n\t%s,\n\t%s,\n\t%s" % (
            self.input,
            self.average_time,
            self.horizon,
            self.initial_trend)


@dataclass
class SampleIfTrueStructure(AbstractSyntax):
    """
    Dataclass for a sample if true structure.

    Parameters
    ----------
    condition: AST
        The condition of the sample if true
    input: AST
        The input of the sample if true.
    initial: AST
        The initial value of the sample if true.

    """
    condition: Union[AbstractSyntax, float]
    input: Union[AbstractSyntax, float]
    initial: Union[AbstractSyntax, float]

    def __str__(self) -> str:  # pragma: no cover
        return "SampleIfTrueStructure:\n\t%s,\n\t%s,\n\t%s" % (
            self.condition,
            self.input,
            self.initial)


@dataclass
class LookupsStructure(AbstractSyntax):
    """
    Dataclass for a lookup structure.

    Parameters
    ----------
    x: tuple
        The list of the x values of the lookup.
    y: tuple
        The list of the y values of the lookup.
    x_limits: tuple
        The minimum and maximum value of x.
    y_limits: tuple
        The minimum and maximum value of y.
    type: str
        The interpolation method.

    """
    x: tuple
    y: tuple
    x_limits: tuple
    y_limits: tuple
    type: str

    def __str__(self) -> str:  # pragma: no cover
        return "LookupStructure (%s):\n\tx %s = %s\n\ty %s = %s\n" % (
            self.type, self.x_limits, self.x, self.y_limits, self.y
        )


@dataclass
class InlineLookupsStructure(AbstractSyntax):
    """
    Dataclass for an inline lookup structure.

    Parameters
    ----------
    argument: AST
        The argument of the inline lookup.
    lookups: LookupStructure
        The lookups definition.

    """
    argument: Union[AbstractSyntax, float]
    lookups: LookupsStructure

    def __str__(self) -> str:  # pragma: no cover
        return "InlineLookupsStructure:\n\t%s\n\t%s" % (
            str(self.argument).replace("\n", "\n\t"),
            str(self.lookups).replace("\n", "\n\t")
        )


@dataclass
class DataStructure(AbstractSyntax):
    """
    Dataclass for an empty data structure.

    Parameters
    ----------
    None

    """
    pass

    def __str__(self) -> str:  # pragma: no cover
        return "DataStructure"


@dataclass
class GetLookupsStructure(AbstractSyntax):
    """
    Dataclass for a get lookups structure.

    Parameters
    ----------
    file: str
        The file path where the data is.
    tab: str
        Tab name from which the data is read. If file type is not a
        spreadsheet this will be used as a separator.
    x_row_or_col: str
        The pointer to the cell or cellrange name that defines the
        interpolation series data.
    cell: str
        The pointer to the cell or the cellrange name that defines the data.

    """
    file: str
    tab: str
    x_row_or_col: str
    cell: str

    def __str__(self) -> str:  # pragma: no cover
        return "GetLookupStructure:\n\t'%s', '%s', '%s', '%s'\n" % (
            self.file, self.tab, self.x_row_or_col, self.cell
        )


@dataclass
class GetDataStructure(AbstractSyntax):
    """
    Dataclass for a get lookups structure.

    Parameters
    ----------
    file: str
        The file path where the data is.
    tab: str
        Tab name from which the data is read. If file type is not a
        spreadsheet this will be used as a separator.
    time_row_or_col: str
        The pointer to the cell or cellrange name that defines the
        interpolation time series data.
    cell: str
        The pointer to the cell or the cellrange name that defines the data.

    """
    file: str
    tab: str
    time_row_or_col: str
    cell: str

    def __str__(self) -> str:  # pragma: no cover
        return "GetDataStructure:\n\t'%s', '%s', '%s', '%s'\n" % (
            self.file, self.tab, self.time_row_or_col, self.cell
        )


@dataclass
class GetConstantsStructure(AbstractSyntax):
    """
    Dataclass for a get lookups structure.

    Parameters
    ----------
    file: str
        The file path where the data is.
    tab: str
        Tab name from which the data is read. If file type is not a
        spreadsheet this will be used as a separator.
    cell: str
        The pointer to the cell or the cellrange name that defines the data.

    """
    file: str
    tab: str
    cell: str

    def __str__(self) -> str:  # pragma: no cover
        return "GetConstantsStructure:\n\t'%s', '%s', '%s'\n" % (
            self.file, self.tab, self.cell
        )
